# :coding: utf-8

import copy
import collections
import base64
import hashlib

import pytest
import six

import wiz.definition
import wiz.package
import wiz.utility
from wiz.utility import Requirement


@pytest.fixture()
def mocked_extract_version_ranges(mocker):
    """Return mocked wiz.utility.extract_version_ranges function."""
    return mocker.patch.object(wiz.utility, "extract_version_ranges")


@pytest.fixture()
def mocked_match(mocker):
    """Return mocked wiz.utility.match function."""
    return mocker.patch.object(wiz.utility, "match")


@pytest.mark.parametrize("element", [
    "This is a string",
    42,
    ["This", "is", "a", "list"],
    {"key": "value"},
    {
        "key1": ["This", "is", "a", "list", 42],
        "key2": {"test": "This is a test"},
        "key3": 1337
    }
], ids=[
    "string",
    "number",
    "list",
    "dict",
    "complex-dict",
])
def test_encode_and_decode(element):
    """Encode *element* and immediately decode it."""
    encoded = wiz.utility.encode(element)
    assert isinstance(encoded, six.string_types)
    assert element == wiz.utility.decode(encoded)


@pytest.mark.parametrize("ranges1, ranges2, expected", [
    (
        [(None, None)],
        [(None, None)],
        True
    ),
    (
        [(None, None)],
        [((1, 1), (4,))],
        True
    ),
    (
        [((1, 1), (4,))],
        [(None, None)],
        True
    ),
    (
        [((1, 1), (4,))],
        [((1, 1), (4,))],
        True
    ),
    (
        [((4,), (4,))],
        [((1, 1), (4,))],
        True
    ),
    (
        [((5,), (7,))],
        [((4,), (4,))],
        False
    ),
    (
        [((4,), (4,))],
        [((5,), (7,))],
        False
    )
], ids=[
    "yes-no-boundaries",
    "yes-no-boundaries-left",
    "yes-no-boundaries-right",
    "yes-same-ranges",
    "yes-one-version-in",
    "no-out-of-scope-left",
    "no-out-of-scope-right"

])
def test_is_overlapping(
    mocker, mocked_extract_version_ranges, ranges1, ranges2, expected
):
    """Indicates whether requirements are overlapping."""
    req1 = mocker.Mock(extras=set())
    req2 = mocker.Mock(extras=set())

    # Hack due to the impossibility to directly mock an attribute called "name"
    # See: https://bradmontgomery.net/blog/how-world-do-you-mock-name-attribute
    type(req1).name = mocker.PropertyMock(return_value="foo")
    type(req2).name = mocker.PropertyMock(return_value="foo")

    mocked_extract_version_ranges.side_effect = [ranges1, ranges2]
    assert wiz.utility.is_overlapping(req1, req2) == expected


def test_is_overlapping_with_variants(mocker):
    """Fail to compare requirement with different variant."""
    req1 = mocker.Mock(extras={"V1"})
    req2 = mocker.Mock(extras={"V2"})

    # Hack due to the impossibility to directly mock an attribute called "name"
    # See: https://bradmontgomery.net/blog/how-world-do-you-mock-name-attribute
    type(req1).name = mocker.PropertyMock(return_value="foo")
    type(req2).name = mocker.PropertyMock(return_value="foo")

    assert wiz.utility.is_overlapping(req1, req2) is False


def test_is_overlapping_fail(mocker):
    """Fail to compare requirement with different name."""
    req1 = mocker.Mock(extras=set())
    req2 = mocker.Mock(extras=set())

    # Hack due to the impossibility to directly mock an attribute called "name"
    # See: https://bradmontgomery.net/blog/how-world-do-you-mock-name-attribute
    type(req1).name = mocker.PropertyMock(return_value="foo")
    type(req2).name = mocker.PropertyMock(return_value="bar")

    with pytest.raises(ValueError) as error:
        wiz.utility.is_overlapping(req1, req2)

    assert (
        "Impossible to compare requirements with different names "
        "['foo' and 'bar']."
    ) in str(error)


@pytest.mark.parametrize("requirement, expected", [
    (
        Requirement("foo"),
        [(None, None)]
    ),
    (
        Requirement("foo <=0.1.0"),
        [(None, (0, 1, 0))]
    ),
    (
        Requirement("foo >=0.1.0"),
        [((0, 1, 0), None)]
    ),
    (
        Requirement("foo >=0.1.0, <=1"),
        [((0, 1, 0), (1,))]
    ),
    (
        Requirement("foo <0.1.0"),
        [(None, (0, 0, 9999))]
    ),
    (
        Requirement("foo >0.1.0"),
        [((0, 1, 0, 1), None)]
    ),
    (
        Requirement("foo >0.1.0, <1"),
        [((0, 1, 0, 1), (0, 9999,))]
    ),
    (
        Requirement("foo >=0.1.0, <1"),
        [((0, 1, 0), (0, 9999,))]
    ),
    (
        Requirement("foo ==0.1.0"),
        [((0, 1, 0), (0, 1, 0))]
    ),
    (
        Requirement("foo ==0.1.*"),
        [((0, 1), (0, 1, 9999))]
    ),
    (
        Requirement("foo ~=0.5"),
        [((0, 5), (0, 5, 9999))]
    ),
    (
        Requirement("foo !=0.3.9"),
        [(None, (0, 3, 8, 9999)), ((0, 3, 9, 1), None)]
    ),
    (
        Requirement("foo !=0.3.*"),
        [(None, (0, 2, 9999)), ((0, 4), None)]
    ),
    (
        Requirement("foo >=9, >3, >10, <=100, <19, !=11.*, !=11.0.*"),
        [((10, 1), (10, 9999)), ((12,), (18, 9999))]
    ),
    (
        Requirement("foo !=1.*, !=1.0.0, <2, ==0.1.0"),
        [((0, 1, 0), (0, 1, 0))]
    )
], ids=[
    "none",
    "inclusive-comparison",
    "inclusive-comparison-min",
    "inclusive-comparison",
    "exclusive-comparison-max",
    "exclusive-comparison-min",
    "exclusive-comparison",
    "mixed-comparison",
    "version-matching",
    "version-matching-with-wildcard",
    "compatible-release",
    "version-exclusion",
    "version-exclusion-with-wildcard",
    "mixed-1",
    "mixed-2"
])
def test_extract_version_ranges(requirement, expected):
    """Extract version ranges from requirements."""
    assert wiz.utility.extract_version_ranges(requirement) == expected


@pytest.mark.parametrize("requirement, expected", [
    (
        Requirement("foo ===8"),
        "Operator '===' is not accepted for requirement 'foo ===8'"
    ),
    (
        Requirement("foo >=9, <8"),
        "The requirement is incorrect as minimum value '9' cannot be set "
        "when maximum value is '7.9999'."
    ),
    (
        Requirement("foo >=8, <8"),
        "The requirement is incorrect as minimum value '8' cannot be set "
        "when maximum value is '7.9999'."
    ),
    (
        Requirement("foo ==1, ==2"),
        "The requirement is incorrect as minimum value '2' cannot be set "
        "when maximum value is '1'."
    ),
    (
        Requirement("foo ==1, !=1.*"),
        "The requirement is incorrect as excluded version range '0.9999-2' "
        "makes all other versions unreachable."
    )
], ids=[
    "incorrect-operator",
    "incorrect-comparison-1",
    "incorrect-comparison-2",
    "incorrect-comparison-3",
    "incorrect-exclusion"
])
def test_extract_version_ranges_error(requirement, expected):
    """Fail to extract version ranges from requirements."""
    with pytest.raises(wiz.exception.RequirementError) as error:
        wiz.utility.extract_version_ranges(requirement)

    assert expected in str(error)


@pytest.mark.parametrize("version, ranges, expected", [
    (
        (1,),
        [(None, None)],
        [((1,), None)]
    ),
    (
        (1, 3, 3),
        [((1, 2, 3), (1, 3, 0)), ((1, 3, 3), (1, 4))],
        [((1, 3, 3), (1, 4))]
    ),
    (
        (1, 2, 9),
        [((1, 2, 3), (1, 3, 0)), ((1, 3, 3), (1, 4))],
        [((1, 2, 9), (1, 3, 0)), ((1, 3, 3), (1, 4))],
    ),
    (
        (1,),
        [(None, (0, 9999)), ((2,), None)],
        [((2,), None)]
    ),
    (
        (1,),
        [(None, (1, 1, 0))],
        [((1,), (1, 1, 0))]
    ),
    (
        (1, 0, 0),
        [((1, 2, 3), (1, 3, 0)), ((1, 3, 3), (1, 4))],
        [((1, 2, 3), (1, 3, 0)), ((1, 3, 3), (1, 4))],
    ),
], ids=[
    "minimal",
    "simple-01",
    "simple-02",
    "simple-03",
    "simple-04",
    "unchanged",
])
def test_update_minimum_version(version, ranges, expected):
    """Update range with minimum value."""
    wiz.utility._update_minimum_version(version, ranges)
    assert ranges == expected


def test_update_minimum_version_fail():
    """Fail to update range with minimum value."""
    with pytest.raises(wiz.exception.RequirementError) as error:
        wiz.utility._update_minimum_version((1, 2, 3), [(None, (1,))])

    assert (
        "The requirement is incorrect as minimum value '1.2.3' cannot be set "
        "when maximum value is '1'."
    ) in str(error)


@pytest.mark.parametrize("version, ranges, expected", [
    (
        (1,),
        [(None, None)],
        [(None, (1,))]
    ),
    (
        (1, 3, 3),
        [((1, 2, 3), (1, 3, 0)), ((1, 3, 3), (1, 4))],
        [((1, 2, 3), (1, 3, 0)), ((1, 3, 3), (1, 3, 3))]
    ),
    (
        (1, 2, 9),
        [((1, 2, 3), (1, 3, 0)), ((1, 3, 3), (1, 4))],
        [((1, 2, 3), (1, 2, 9))],
    ),
    (
        (1,),
        [(None, (0, 9999)), ((2,), None)],
        [(None, (0, 9999))]
    ),
    (
        (1,),
        [((0, 1, 0), None)],
        [((0, 1, 0), (1,))]
    ),
    (
        (2, 0, 0),
        [((1, 2, 3), (1, 3, 0)), ((1, 3, 3), (1, 4))],
        [((1, 2, 3), (1, 3, 0)), ((1, 3, 3), (1, 4))],
    ),
], ids=[
    "minimal",
    "simple-01",
    "simple-02",
    "simple-03",
    "simple-04",
    "unchanged",
])
def test_update_maximum_version(version, ranges, expected):
    """Update range with minimum value."""
    wiz.utility._update_maximum_version(version, ranges)
    assert ranges == expected


def test_update_maximum_version_fail():
    """Fail to update range with maximum value."""
    with pytest.raises(wiz.exception.RequirementError) as error:
        wiz.utility._update_maximum_version((1, 2, 3), [((2,), None)])

    assert (
        "The requirement is incorrect as maximum value '1.2.3' cannot be set "
        "when minimum value is '2'."
    ) in str(error)


@pytest.mark.parametrize("excluded, ranges, expected", [
    (
        ((1,), (2,)),
        [(None, None)],
        [(None, (1,)), ((2,), None)]
    ),
    (
        ((1,), (2,)),
        [((0,), (3,))],
        [((0,), (1,)), ((2,), (3,))]
    ),
    (
        ((1,), (2,)),
        [((1,), (2,))],
        [((1,), (1,)), ((2,), (2,))],
    ),
    (
        ((1,), (2,)),
        [(None, (1, 2, 3))],
        [(None, (1,))]
    ),
    (
        ((1,), (2,)),
        [((1, 2, 3), None)],
        [((2,), None)]
    ),
    (
        ((1,), (2,)),
        [((2,), (3,))],
        [((2,), (3,))],
    ),
    (
        ((4,), (5,)),
        [((2,), (3,))],
        [((2,), (3,))],
    ),
], ids=[
    "exclude-middle-range-01",
    "exclude-middle-range-02",
    "exclude-middle-range-03",
    "exclude-end-range",
    "exclude-start-range",
    "out-of-range-01",
    "out-of-range-02",
])
def test_update_version_ranges(excluded, ranges, expected):
    """Update version ranges from excluded version range."""
    wiz.utility._update_version_ranges(excluded, ranges)
    assert ranges == expected


def test_update_version_ranges_fail():
    """Fail to update version ranges from excluded version range."""
    with pytest.raises(wiz.exception.RequirementError) as error:
        wiz.utility._update_version_ranges(((0,), (3,)), [((1,), (2,))])

    assert (
        "The requirement is incorrect as excluded version range '0-3' "
        "makes all other versions unreachable."
    ) in str(error)


@pytest.mark.parametrize("definition, expected", [
    (
        wiz.definition.Definition({"identifier": "test"}),
        "'test'"
    ),
    (
        wiz.definition.Definition({
            "identifier": "test",
            "version": "0.1.0"
        }),
        "'test' [0.1.0]"
    ),
    (
        wiz.definition.Definition({
            "identifier": "test",
            "namespace": "foo"
        }),
        "'foo::test'"
    ),
    (
        wiz.definition.Definition({
            "identifier": "test",
            "system": {
                "platform": "linux"
            }
        }),
        "'test' (linux)"
    ),
    (
        wiz.definition.Definition({
            "identifier": "test",
            "version": "0.1.0",
            "namespace": "foo",
            "system": {
                "platform": "linux"
            }
        }),
        "'foo::test' [0.1.0] (linux)"
    )
], ids=[
    "simple",
    "with-version",
    "with-namespace",
    "with-system",
    "with-all",
])
def test_compute_label(definition, expected):
    """Compute definition label."""
    assert wiz.utility.compute_label(definition) == expected


@pytest.mark.parametrize("definition, expected", [
    (
        wiz.definition.Definition({"identifier": "test"}),
        "test.json"
    ),
    (
        wiz.definition.Definition({
            "identifier": "test",
            "version": "0.1.0"
        }),
        "test-0.1.0.json"
    ),
    (
        wiz.definition.Definition({
            "identifier": "test",
            "namespace": "foo"
        }),
        "foo-test.json"
    ),
    (
        wiz.definition.Definition({
            "identifier": "test",
            "namespace": "foo::bar::bim"
        }),
        "foo-bar-bim-test.json"
    ),
    (
        wiz.definition.Definition({
            "identifier": "test",
            "system": {
                "platform": "linux"
            }
        }),
        "test-{}.json".format(
            base64.urlsafe_b64encode(
                hashlib.sha1(b"linux").digest()
            ).rstrip(b"=").decode("utf-8")
        )
    ),
    (
        wiz.definition.Definition({
            "identifier": "test",
            "version": "0.1.0",
            "namespace": "bar::foo",
            "system": {
                "platform": "linux"
            }
        }),
        "bar-foo-test-0.1.0-{}.json".format(
            base64.urlsafe_b64encode(
                hashlib.sha1(b"linux").digest()
            ).rstrip(b"=").decode("utf-8")
        )
    )
], ids=[
    "simple",
    "with-version",
    "with-namespace",
    "with-multiple-namespaces",
    "with-system",
    "with-all",
])
def test_compute_file_name(definition, expected):
    """Compute definition file name."""
    assert wiz.utility.compute_file_name(definition) == expected


@pytest.mark.parametrize("mapping1, mapping2, expected", [
    ({}, {}, {}),
    ({"A": 1, "B": 2}, {"B": 3}, {"A": 1, "B": 3}),
    (
        {"A": 1, "B": {"C": 3, "D": 4}},
        {"B": {"C": 4}},
        {"A": 1, "B": {"C": 4, "D": 4}},
    ),
    (
        {"A": 1, "B": {"C": {"D": 4, "E": 5}}},
        {"B": {"C": {"E": 6}}},
        {"A": 1, "B": {"C": {"D": 4, "E": 6}}},
    ),
    (
        {"A": 1, "B": {"C": {"D": {"E": {"F": {"G": 2, "H": 3}}}}}},
        {"B": {"C": {"D": {"E": {"F": {"H": 4, "I": 5}}}}}},
        {"A": 1, "B": {"C": {"D": {"E": {"F": {"G": 2, "H": 4, "I": 5}}}}}},
    )
], ids=[
    "empty",
    "simple",
    "one-level-deep",
    "two-level-deep",
    "five-level-deep"
])
def test_deep_update(mapping1, mapping2, expected):
    """Recursively update mapping."""
    _mapping2 = copy.deepcopy(mapping2)
    assert wiz.utility.deep_update(mapping1, mapping2) == expected
    assert mapping1 == expected
    assert mapping2 == _mapping2


@pytest.mark.parametrize("package_data, variant_index, requirement, expected", [
    ({"identifier": "A"}, None, Requirement("A"), Requirement("::A")),
    (
        {"identifier": "A", "version": "2"}, None,
        Requirement("A > 1"), Requirement("::A > 1")
    ),
    (
        {"identifier": "A", "namespace": "foo"}, None,
        Requirement("A"), Requirement("foo::A")
    ),
    (
        {"identifier": "A", "namespace": "foo::bar"}, None,
        Requirement("A"), Requirement("foo::bar::A")
    ),
    (
        {"identifier": "A", "variants": [{"identifier": "V1"}]}, 0,
        Requirement("A[V1]"), Requirement("::A[V1]")
    )
], ids=[
    "simple",
    "with-version",
    "with-namespace",
    "with-long-namespace",
    "with-variant",
])
def test_sanitize_requirement(
    mocked_match, package_data, variant_index, requirement, expected
):
    """Return qualified requirement depending on package's namespace."""
    mocked_match.return_value = True
    package = wiz.package.Package(
        wiz.definition.Definition(package_data), variant_index=variant_index
    )
    assert wiz.utility.sanitize_requirement(requirement, package) == expected


def test_sanitize_requirement_error(mocked_match):
    """Fail to return qualified requirement."""
    mocked_match.return_value = False
    package = wiz.package.Package(
        wiz.definition.Definition({"identifier": "B"})
    )

    with pytest.raises(ValueError) as error:
        wiz.utility.sanitize_requirement(Requirement("A"), package)

    assert "Requirement 'A' is incompatible with package 'B'" in str(error)


@pytest.mark.parametrize("requirements, expected", [
    ([], {}),
    ([Requirement("foo"), Requirement("bar")], {"A": 1, "B": 1, "C": 1}),
    ([Requirement("::foo"), Requirement("B::bar")], {"B": 1}),
    ([Requirement("foo"), Requirement("B::bar")], {"A": 1, "B": 1}),
], ids=[
    "empty",
    "non-qualified-requirements",
    "qualified-requirements",
    "mixed",
])
def test_namespace_counter(requirements, expected):
    """Compute namespace occurrences counter from requirements."""
    mapping = {
        "__namespace__": {
            "foo": ["A"],
            "bar": ["B", "C"]
        }
    }
    result = wiz.utility.compute_namespace_counter(requirements, mapping)
    assert result == collections.Counter(expected)


@pytest.mark.parametrize("package_data, variant_index, requirement, expected", [
    ({"identifier": "A"}, None, Requirement("A"), True),
    ({"identifier": "B"}, None, Requirement("A"), False),
    ({"identifier": "A", "version": "1"}, None, Requirement("A>=1"), True),
    ({"identifier": "A", "version": "0.1.0"}, None, Requirement("A>=1"), False),
    ({"identifier": "A", "namespace": "ba"}, None, Requirement("A"), True),
    ({"identifier": "A", "namespace": "ba"}, None, Requirement("ba::A"), True),
    ({"identifier": "A", "namespace": "to"}, None, Requirement("ba::A"), False),
    (
        {"identifier": "A", "variants": [{"identifier": "V1"}]}, 0,
        Requirement("A"), True
    ),
    (
        {"identifier": "A", "variants": [{"identifier": "V1"}]}, 0,
        Requirement("A[V1]"), True
    ),
    (
        {"identifier": "A", "variants": [{"identifier": "V1"}]}, 0,
        Requirement("A[V3]"), False
    ),
], ids=[
    "matching-identifier",
    "non-matching-identifier",
    "matching-version",
    "non-matching-version",
    "matching-namespace1",
    "matching-namespace1",
    "non-matching-namespace",
    "matching-variant1",
    "matching-variant2",
    "non-matching-variant",
])
def test_match(package_data, variant_index, requirement, expected):
    """Return whether requirement is compatible with package."""
    package = wiz.package.Package(
        wiz.definition.Definition(package_data), variant_index=variant_index
    )
    assert wiz.utility.match(requirement, package) == expected


@pytest.mark.parametrize("requirement, namespace, identifier", [
    (Requirement("A"), None, "A"),
    (Requirement("::A"), None, "A"),
    (Requirement("foo::A"), "foo", "A"),
    (Requirement("foo::bar::A"), "foo::bar", "A"),
], ids=[
    "without-namespace",
    "qualified-without-namespace",
    "with-namespace",
    "with-long-namespace",
])
def test_extract_namespace(requirement, namespace, identifier):
    """Extract namespace and identifier from requirement."""
    assert wiz.utility.extract_namespace(requirement) == (namespace, identifier)


@pytest.mark.parametrize("requirements1, requirements2, expected", [
    ([], [], False),
    (["bim > 3", "bah"], ["zim"], False),
    (["bim > 3", "bah"], ["bim"], False),
    (["bim > 3", "bah"], ["bim < 3"], True),
], ids=[
    "without-requirements",
    "different-requirements",
    "compatible-requirements",
    "incompatible-requirements",
])
def test_check_conflicting_requirements(requirements1, requirements2, expected):
    """Check whether some requirements are conflicting between packages."""
    data1 = {"identifier": "foo"}
    if len(requirements1):
        data1["requirements"] = requirements1

    data2 = {"identifier": "bar"}
    if len(requirements2):
        data2["requirements"] = requirements2

    package1 = wiz.package.Package(wiz.definition.Definition(data1))
    package2 = wiz.package.Package(wiz.definition.Definition(data2))

    result = wiz.utility.check_conflicting_requirements(package1, package2)
    assert result == expected



# :coding: utf-8

import wiz.command_line


def main():
    """Execute main command line interface passing command line arguments."""
    wiz.command_line.main(prog_name="wiz")


if __name__ == "__main__":
    main()



# :coding: utf-8

import copy
import os

import pytest
from packaging.requirements import Requirement

import wiz
import wiz.definition
import wiz.environ
import wiz.exception
import wiz.graph
import wiz.package
import wiz.system
import wiz.utility
from wiz._version import __version__


@pytest.fixture()
def mocked_registry_defaults(mocker):
    """Return mocked 'wiz.registry.get_default' function."""
    return mocker.patch.object(wiz.registry, "get_defaults")


@pytest.fixture()
def mocked_registry_install_to_path(mocker):
    """Return mocked 'wiz.registry.install_to_path' function."""
    return mocker.patch.object(wiz.registry, "install_to_path")


@pytest.fixture()
def mocked_fetch_definition_mapping(mocker):
    """Return mocked 'wiz.fetch_definition_mapping' function."""
    return mocker.patch.object(wiz, "fetch_definition_mapping")


@pytest.fixture()
def mocked_fetch_package(mocker):
    """Return mocked 'wiz.fetch_package' function."""
    return mocker.patch.object(wiz, "fetch_package")


@pytest.fixture()
def mocked_load_definition(mocker):
    """Return mocked 'wiz.load_definition' function."""
    return mocker.patch.object(wiz, "load_definition")


@pytest.fixture()
def mocked_definition_fetch(mocker):
    """Return mocked 'wiz.definition.fetch' function."""
    return mocker.patch.object(wiz.definition, "fetch")


@pytest.fixture()
def mocked_definition_query(mocker):
    """Return mocked 'wiz.definition.query' function."""
    return mocker.patch.object(wiz.definition, "query")


@pytest.fixture()
def mocked_definition_load(mocker):
    """Return mocked 'wiz.definition.load' function."""
    return mocker.patch.object(wiz.definition, "load")


@pytest.fixture()
def mocked_definition_export(mocker):
    """Return mocked 'wiz.definition.export' function."""
    return mocker.patch.object(wiz.definition, "export")


@pytest.fixture()
def mocked_package_extract(mocker):
    """Return mocked 'wiz.package.extract' function."""
    return mocker.patch.object(wiz.package, "extract")


@pytest.fixture()
def mocked_environ_initiate(mocker):
    """Return mocked 'wiz.environ.initiate' function."""
    return mocker.patch.object(wiz.environ, "initiate")


@pytest.fixture()
def mocked_package_extract_context(mocker):
    """Return mocked 'wiz.package.extract_context' function."""
    return mocker.patch.object(wiz.package, "extract_context")


@pytest.fixture()
def mocked_system_query(mocker):
    """Return mocked 'wiz.definition.fetch' function."""
    return mocker.patch.object(wiz.system, "query")


@pytest.fixture()
def mocked_graph_resolver(mocker):
    """Return mocked 'wiz.graph.Resolver' class constructor."""
    return mocker.patch.object(wiz.graph, "Resolver")


@pytest.fixture()
def mocked_utility_encode(mocker):
    """Return mocked 'wiz.utility.encode' function."""
    return mocker.patch.object(wiz.utility, "encode")


@pytest.fixture()
def mocked_utility_decode(mocker):
    """Return mocked 'wiz.utility.decode' function."""
    return mocker.patch.object(wiz.utility, "decode")


@pytest.fixture()
def mocked_compute_namespace_counter(mocker):
    """Return mocked 'wiz.utility.compute_namespace_counter' function.

    Ensure that 'requirements' list is copied to better analyze state
    evolution.

    .. seealso::

        https://docs.python.org/3/library/unittest.mock-examples.html
        #coping-with-mutable-arguments

    """
    class _MagicMock(mocker.MagicMock):
        """Extended mocker to copy namespace counter on each call."""

        def __call__(self, *args, **kwargs):
            args = copy.deepcopy(args)
            return super(_MagicMock, self).__call__(*args, **kwargs)

    return mocker.patch.object(
        wiz.utility, "compute_namespace_counter", _MagicMock()
    )


@pytest.mark.parametrize("options", [
    {},
    {"max_depth": 2},
    {"system_mapping": "__CUSTOM_SYSTEM_MAPPING__"}
], ids=[
    "paths-only",
    "with-max-depth",
    "with-system-mapping",
])
def test_fetch_definition_mapping(
    mocked_definition_fetch, mocked_system_query, options
):
    """Fetch mapping from all definitions available."""
    paths = ["/path/to/registry1", "/path/to/registry2"]
    default_system_mapping = "__SYSTEM_MAPPING__"
    definition_mapping = {"command": {}, "package": {}}

    mocked_system_query.return_value = default_system_mapping
    mocked_definition_fetch.return_value = definition_mapping

    result = wiz.fetch_definition_mapping(paths, **options)

    definition_mapping.update({"registries": paths})
    assert result == definition_mapping

    mocked_definition_fetch.assert_called_once_with(
        paths,
        max_depth=options.get("max_depth"),
        system_mapping=options.get("system_mapping", default_system_mapping)
    )

    if options.get("system_mapping"):
        mocked_system_query.assert_not_called()
    else:
        mocked_system_query.assert_called_once()


def test_fetch_definition(mocked_definition_query):
    """Fetch definition."""
    request = "test >= 10"
    definition_mapping = {
        "command": "__COMMAND_MAPPING__",
        "package": "__PACKAGE_MAPPING__"
    }

    wiz.fetch_definition(request, definition_mapping)
    mocked_definition_query.assert_called_once_with(
        Requirement(request), "__PACKAGE_MAPPING__"
    )


def test_fetch_package(mocked_package_extract):
    """Fetch package."""
    request = "test[variant] >= 10"
    definition_mapping = {
        "command": "__COMMAND_MAPPING__",
        "package": "__PACKAGE_MAPPING__"
    }

    wiz.fetch_package(request, definition_mapping)
    mocked_package_extract.assert_called_once_with(
        Requirement(request), "__PACKAGE_MAPPING__"
    )


def test_fetch_package_request_from_command():
    """Fetch package request corresponding to command."""
    definition_mapping = {
        "command": {"app": "test"},
        "package": {
            "test": {
                "0.1.0": "__PACKAGE_DEFINITION__"
            }
        }
    }

    assert wiz.fetch_package_request_from_command(
        "app", definition_mapping
    ) == "test"

    assert wiz.fetch_package_request_from_command(
        "app >=10", definition_mapping
    ) == "test >=10"

    assert wiz.fetch_package_request_from_command(
        "app[variant]", definition_mapping
    ) == "test[variant]"

    assert wiz.fetch_package_request_from_command(
        "app[variant] ==10.0.*", definition_mapping
    ) == "test[variant] ==10.0.*"

    assert wiz.fetch_package_request_from_command(
        "app[variant] >1, <2", definition_mapping
    ) == "test[variant] >1, <2"


def test_fetch_package_request_from_command_error():
    """Fail to fetch package request corresponding to command."""
    definition_mapping = {
        "command": {},
    }

    with pytest.raises(wiz.exception.RequestNotFound):
        wiz.fetch_package_request_from_command("app", definition_mapping)


@pytest.mark.parametrize("options, environ, max_combinations, max_attempts", [
    ({}, None, None, None),
    ({"environ_mapping": "__ENVIRON__"}, "__ENVIRON__", None, None),
    ({"maximum_combinations": 1}, None, 1, None),
    ({"maximum_attempts": 1}, None, None, 1),
], ids=[
    "simple",
    "with-environ",
    "with-maximum-combinations",
    "with-maximum-attempts",
])
def test_resolve_context(
    mocked_registry_defaults, mocked_fetch_definition_mapping,
    mocked_graph_resolver, mocked_environ_initiate,
    mocked_package_extract_context, mocked_utility_encode,
    mocked_compute_namespace_counter, mocker, options, environ,
    max_combinations, max_attempts
):
    """Get resolved context mapping."""
    requests = ["test1 >=10, < 11", "test2", "test3[variant]"]
    paths = ["/path/to/registry1", "/path/to/registry2"]

    context = {"environ": {"KEY": "VALUE"}, "command": {"app": "APP"}}
    packages = [
        mocker.Mock(identifier="test1"),
        mocker.Mock(identifier="test2"),
        mocker.Mock(identifier="test3")
    ]

    mocked_resolver = mocker.Mock(**{"compute_packages.return_value": packages})
    mocked_graph_resolver.return_value = mocked_resolver
    mocked_environ_initiate.return_value = "__INITIAL_ENVIRON__"
    mocked_package_extract_context.return_value = context
    mocked_utility_encode.return_value = "__ENCODED_CONTEXT__"

    definition_mapping = {
        "package": "__PACKAGE_DEFINITIONS__",
        "registries": paths
    }

    result = wiz.resolve_context(requests, definition_mapping, **options)

    assert result == {
        "environ": {
            "KEY": "VALUE",
            "WIZ_VERSION": __version__,
            "WIZ_CONTEXT": "__ENCODED_CONTEXT__"
        },
        "command": {"app": "APP"},
        "packages": packages,
        "registries": paths
    }

    mocked_registry_defaults.assert_not_called()
    mocked_fetch_definition_mapping.assert_not_called()

    mocked_compute_namespace_counter.assert_called_once_with(
        [Requirement(request) for request in requests],
        "__PACKAGE_DEFINITIONS__",
    )

    mocked_graph_resolver.assert_called_once_with(
        "__PACKAGE_DEFINITIONS__",
        maximum_combinations=max_combinations,
        maximum_attempts=max_attempts
    )
    mocked_resolver.compute_packages.assert_called_once_with(
        [Requirement(request) for request in requests],
        namespace_counter=mocked_compute_namespace_counter.return_value
    )
    mocked_environ_initiate.assert_called_once_with(environ)

    mocked_package_extract_context.assert_called_once_with(
        packages, environ_mapping="__INITIAL_ENVIRON__"
    )


@pytest.mark.parametrize("options, environ, max_combinations, max_attempts", [
    ({}, None, None, None),
    ({"environ_mapping": "__ENVIRON__"}, "__ENVIRON__", None, None),
    ({"maximum_combinations": 1}, None, 1, None),
    ({"maximum_attempts": 1}, None, None, 1),
], ids=[
    "simple",
    "with-environ",
    "with-maximum-combinations",
    "with-maximum-attempts",
])
def test_resolve_context_with_default_definition_mapping(
    mocked_registry_defaults, mocked_fetch_definition_mapping,
    mocked_graph_resolver, mocked_environ_initiate,
    mocked_package_extract_context, mocked_utility_encode,
    mocked_compute_namespace_counter, mocker, options, environ,
    max_combinations, max_attempts
):
    """Get resolved context mapping with default definition mapping."""
    requests = ["test1 >=10, < 11", "test2", "test3[variant]"]
    paths = ["/path/to/registry1", "/path/to/registry2"]

    context = {"environ": {"KEY": "VALUE"}, "command": {"app": "APP"}}
    packages = [
        mocker.Mock(identifier="test1"),
        mocker.Mock(identifier="test2"),
        mocker.Mock(identifier="test3")
    ]

    mocked_resolver = mocker.Mock(**{"compute_packages.return_value": packages})
    mocked_graph_resolver.return_value = mocked_resolver
    mocked_environ_initiate.return_value = "__INITIAL_ENVIRON__"
    mocked_package_extract_context.return_value = context
    mocked_utility_encode.return_value = "__ENCODED_CONTEXT__"
    mocked_registry_defaults.return_value = paths
    mocked_fetch_definition_mapping.return_value = {
        "package": "__PACKAGE_DEFINITIONS__",
        "registries": paths
    }

    result = wiz.resolve_context(requests, **options)

    assert result == {
        "environ": {
            "KEY": "VALUE",
            "WIZ_VERSION": __version__,
            "WIZ_CONTEXT": "__ENCODED_CONTEXT__"
        },
        "command": {"app": "APP"},
        "packages": packages,
        "registries": paths
    }

    mocked_registry_defaults.asset_called_once()
    mocked_fetch_definition_mapping.asset_called_once_with(paths)

    mocked_compute_namespace_counter.assert_called_once_with(
        [Requirement(request) for request in requests],
        "__PACKAGE_DEFINITIONS__",
    )

    mocked_graph_resolver.assert_called_once_with(
        "__PACKAGE_DEFINITIONS__",
        maximum_combinations=max_combinations,
        maximum_attempts=max_attempts
    )
    mocked_resolver.compute_packages.assert_called_once_with(
        [Requirement(request) for request in requests],
        namespace_counter=mocked_compute_namespace_counter.return_value
    )

    mocked_environ_initiate.assert_called_once_with(environ)

    mocked_package_extract_context.assert_called_once_with(
        packages, environ_mapping="__INITIAL_ENVIRON__"
    )


@pytest.mark.parametrize("options, environ, max_combinations, max_attempts", [
    ({}, None, None, None),
    ({"environ_mapping": "__ENVIRON__"}, "__ENVIRON__", None, None),
    ({"maximum_combinations": 1}, None, 1, None),
    ({"maximum_attempts": 1}, None, None, 1),
], ids=[
    "simple",
    "with-environ",
    "with-maximum-combinations",
    "with-maximum-attempts",
])
def test_resolve_context_with_implicit_packages(
    mocked_registry_defaults, mocked_fetch_definition_mapping,
    mocked_graph_resolver, mocked_environ_initiate,
    mocked_package_extract_context, mocked_utility_encode,
    mocked_compute_namespace_counter, mocker, options, environ,
    max_combinations, max_attempts
):
    """Get resolved context mapping with implicit packages."""
    requests = ["test1 >=10, < 11", "test2", "test3[variant]"]
    implicit = ["foo==0.1.0", "bar==5.2.1"]
    paths = ["/path/to/registry1", "/path/to/registry2"]

    context = {"environ": {"KEY": "VALUE"}, "command": {"app": "APP"}}
    packages = [
        mocker.Mock(identifier="test1"),
        mocker.Mock(identifier="test2"),
        mocker.Mock(identifier="test3")
    ]

    mocked_resolver = mocker.Mock(**{"compute_packages.return_value": packages})
    mocked_graph_resolver.return_value = mocked_resolver
    mocked_environ_initiate.return_value = "__INITIAL_ENVIRON__"
    mocked_package_extract_context.return_value = context
    mocked_utility_encode.return_value = "__ENCODED_CONTEXT__"

    definition_mapping = {
        "package": "__PACKAGE_DEFINITIONS__",
        "registries": paths,
        "implicit-packages": implicit
    }

    result = wiz.resolve_context(requests, definition_mapping, **options)

    assert result == {
        "environ": {
            "KEY": "VALUE",
            "WIZ_VERSION": __version__,
            "WIZ_CONTEXT": "__ENCODED_CONTEXT__"
        },
        "command": {"app": "APP"},
        "packages": packages,
        "registries": paths
    }

    mocked_registry_defaults.asset_called_once()
    mocked_fetch_definition_mapping.asset_called_once_with(paths)

    mocked_compute_namespace_counter.assert_called_once_with(
        [Requirement(request) for request in requests],
        "__PACKAGE_DEFINITIONS__",
    )

    mocked_graph_resolver.assert_called_once_with(
        "__PACKAGE_DEFINITIONS__",
        maximum_combinations=max_combinations,
        maximum_attempts=max_attempts
    )
    mocked_resolver.compute_packages.assert_called_once_with(
        [Requirement(request) for request in implicit + requests],
        namespace_counter=mocked_compute_namespace_counter.return_value
    )

    mocked_environ_initiate.assert_called_once_with(environ)

    mocked_package_extract_context.assert_called_once_with(
        packages, environ_mapping="__INITIAL_ENVIRON__"
    )


@pytest.mark.parametrize("options, environ, max_combinations, max_attempts", [
    ({}, None, None, None),
    ({"environ_mapping": "__ENVIRON__"}, "__ENVIRON__", None, None),
    ({"maximum_combinations": 1}, None, 1, None),
    ({"maximum_attempts": 1}, None, None, 1),
], ids=[
    "simple",
    "with-environ",
    "with-maximum-combinations",
    "with-maximum-attempts",
])
def test_resolve_context_with_implicit_packages_ignored(
    mocked_registry_defaults, mocked_fetch_definition_mapping,
    mocked_graph_resolver, mocked_environ_initiate,
    mocked_package_extract_context, mocked_utility_encode,
    mocked_compute_namespace_counter, mocker, options, environ,
    max_combinations, max_attempts
):
    """Get resolved context mapping with implicit packages ignored."""
    requests = ["test1 >=10, < 11", "test2", "test3[variant]"]
    implicit = ["foo==0.1.0", "bar==5.2.1"]
    paths = ["/path/to/registry1", "/path/to/registry2"]

    context = {"environ": {"KEY": "VALUE"}, "command": {"app": "APP"}}
    packages = [
        mocker.Mock(identifier="test1"),
        mocker.Mock(identifier="test2"),
        mocker.Mock(identifier="test3")
    ]

    mocked_resolver = mocker.Mock(**{"compute_packages.return_value": packages})
    mocked_graph_resolver.return_value = mocked_resolver
    mocked_environ_initiate.return_value = "__INITIAL_ENVIRON__"
    mocked_package_extract_context.return_value = context
    mocked_utility_encode.return_value = "__ENCODED_CONTEXT__"

    definition_mapping = {
        "package": "__PACKAGE_DEFINITIONS__",
        "registries": paths,
        "implicit-packages": implicit
    }

    options["ignore_implicit"] = True
    result = wiz.resolve_context(requests, definition_mapping, **options)

    assert result == {
        "environ": {
            "KEY": "VALUE",
            "WIZ_VERSION": __version__,
            "WIZ_CONTEXT": "__ENCODED_CONTEXT__"
        },
        "command": {"app": "APP"},
        "packages": packages,
        "registries": paths
    }

    mocked_registry_defaults.asset_called_once()
    mocked_fetch_definition_mapping.asset_called_once_with(paths)

    mocked_compute_namespace_counter.assert_called_once_with(
        [Requirement(request) for request in requests],
        "__PACKAGE_DEFINITIONS__",
    )

    mocked_graph_resolver.assert_called_once_with(
        "__PACKAGE_DEFINITIONS__",
        maximum_combinations=max_combinations,
        maximum_attempts=max_attempts
    )
    mocked_resolver.compute_packages.assert_called_once_with(
        [Requirement(request) for request in requests],
        namespace_counter=mocked_compute_namespace_counter.return_value
    )

    mocked_environ_initiate.assert_called_once_with(environ)

    mocked_package_extract_context.assert_called_once_with(
        packages, environ_mapping="__INITIAL_ENVIRON__"
    )


def test_resolve_command():
    """Resolve a command from command mapping."""
    elements = ["app", "--option", "value", "/path/to/script"]
    result = wiz.resolve_command(elements, {})
    assert result == elements

    result = wiz.resolve_command(elements, {"app": "App0.1 --modeX"})
    assert result == [
        "App0.1", "--modeX", "--option", "value", "/path/to/script"
    ]


def test_discover_context(
    monkeypatch, mocked_utility_decode, mocked_fetch_definition_mapping,
    mocked_fetch_package, mocked_environ_initiate,
    mocked_package_extract_context,
):
    """Discover context from environment variable."""
    monkeypatch.setenv("WIZ_CONTEXT", "__CONTEXT__")

    paths = ["/path/to/registry1", "/path/to/registry2"]
    package_identifiers = ["package1==0.1.2", "package2==1.0.2"]
    mocked_utility_decode.return_value = [package_identifiers, paths]
    mocked_fetch_definition_mapping.return_value = "__DEFINITION_MAPPING__"
    mocked_fetch_package.side_effect = [
        {"identifier": "package1==0.1.2"}, {"identifier": "package2==1.0.2"}
    ]
    mocked_environ_initiate.return_value = {"KEY": "VALUE"}
    mocked_package_extract_context.return_value = {
        "command": {},
        "environ": {"KEY": "VALUE"},
    }

    result = wiz.discover_context()
    assert result == {
        "registries": paths,
        "command": {},
        "environ": {"KEY": "VALUE"},
        "packages": [
            {"identifier": "package1==0.1.2"},
            {"identifier": "package2==1.0.2"}
        ]
    }

    mocked_utility_decode.assert_called_once_with("__CONTEXT__")
    mocked_fetch_definition_mapping.assert_called_once_with(paths)

    assert mocked_fetch_package.call_count == 2
    mocked_fetch_package.assert_any_call(
        package_identifiers[0], "__DEFINITION_MAPPING__"
    )
    mocked_fetch_package.assert_any_call(
        package_identifiers[1], "__DEFINITION_MAPPING__"
    )

    mocked_package_extract_context.assert_called_once_with(
        [{"identifier": "package1==0.1.2"}, {"identifier": "package2==1.0.2"}],
        environ_mapping={"KEY": "VALUE"}
    )


def test_discover_context_error(monkeypatch):
    """Fail to discover context from environment variable."""
    monkeypatch.delenv("WIZ_CONTEXT", raising=False)

    with pytest.raises(wiz.exception.RequestNotFound):
        wiz.discover_context()


def test_load_definition(mocked_definition_load):
    """Load a definition."""
    wiz.load_definition("/path/to/definition.json")
    mocked_definition_load.assert_called_once_with("/path/to/definition.json")


def test_export_definition(mocked_definition_export):
    """Export a definition to file."""
    definition_data = {
        "identifier": "foo",
        "description": "Test definition",
        "command": {
            "app": "App0.1",
            "appX": "AppX0.1"
        },
        "environ": {
            "KEY": "VALUE"
        }
    }
    wiz.export_definition("/path/to/output", definition_data)
    mocked_definition_export.assert_called_once_with(
        "/path/to/output", definition_data, overwrite=False
    )


@pytest.fixture()
def packages(request, mocker):
    """Return mocked packages as specified by *request.param*."""
    if request.param is None:
        return

    return [
        mocker.Mock(identifier=identifier) for identifier
        in request.param
    ]


@pytest.mark.parametrize("options, packages, expected", [
    (
        {"environ": {"KEY": "VALUE"}},
        None,
        (
            "#!/bin/tcsh -f\n"
            "setenv KEY \"VALUE\"\n"
        )
    ),
    (
        {"environ": {"PATH": "/path/to/bin"}},
        None,
        (
            "#!/bin/tcsh -f\n"
            "setenv PATH \"/path/to/bin:${PATH}\"\n"
        )
    ),
    (
        {"environ": {"KEY": "VALUE"}, "command": "App0.1"},
        None,
        (
            "#!/bin/tcsh -f\n"
            "setenv KEY \"VALUE\"\n"
            "App0.1 $argv:q\n"
        )
    ),
    (
        {"environ": {"KEY": "VALUE"}},
        ["package1", "package2", "package3"],
        (
            "#!/bin/tcsh -f\n"
            "#\n"
            "# Generated by wiz with the following environments:\n"
            "# - package1\n"
            "# - package2\n"
            "# - package3\n"
            "#\n"
            "setenv KEY \"VALUE\"\n"
        )
    ),
    (
        {"environ": {"KEY": "VALUE"}, "command": "App -x"},
        ["packageA", "packageB"],
        (
            "#!/bin/tcsh -f\n"
            "#\n"
            "# Generated by wiz with the following environments:\n"
            "# - packageA\n"
            "# - packageB\n"
            "#\n"
            "setenv KEY \"VALUE\"\n"
            "App -x $argv:q\n"
        )
    )
], ids=[
    "minimal",
    "with-path",
    "with-command",
    "with-packages",
    "with-command-and-packages",
], indirect=[
    "packages"
])
def test_export_csh_script(temporary_directory, options, packages, expected):
    """Export CSH script."""
    options.update({"packages": packages})

    wiz.export_script(
        temporary_directory, "tcsh", "foo", **options
    )

    file_path = os.path.join(temporary_directory, "foo")
    assert os.path.isfile(file_path) is True

    with open(file_path, "r") as stream:
        assert stream.read() == expected


def test_export_csh_script_environ_error(temporary_directory):
    """Fail to export CSH script with empty environment mapping."""
    with pytest.raises(ValueError) as error:
        wiz.export_script(temporary_directory, "tcsh", "foo", {})

    assert "The environment mapping should not be empty." in str(error)


@pytest.mark.parametrize("options, packages, expected", [
    (
        {"environ": {"KEY": "VALUE"}},
        None,
        (
            "#!/bin/bash\n"
            "export KEY=\"VALUE\"\n"
        )
    ),
    (
        {"environ": {"PATH": "/path/to/bin"}},
        None,
        (
            "#!/bin/bash\n"
            "export PATH=\"/path/to/bin:${PATH}\"\n"
        )
    ),
    (
        {"environ": {"KEY": "VALUE"}, "command": "App0.1"},
        None,
        (
            "#!/bin/bash\n"
            "export KEY=\"VALUE\"\n"
            "App0.1 $@\n"
        )
    ),
    (
        {"environ": {"KEY": "VALUE"}},
        ["package1", "package2", "package3"],
        (
            "#!/bin/bash\n"
            "#\n"
            "# Generated by wiz with the following environments:\n"
            "# - package1\n"
            "# - package2\n"
            "# - package3\n"
            "#\n"
            "export KEY=\"VALUE\"\n"
        )
    ),
    (
        {"environ": {"KEY": "VALUE"}, "command": "App -x"},
        ["packageA", "packageB"],
        (
            "#!/bin/bash\n"
            "#\n"
            "# Generated by wiz with the following environments:\n"
            "# - packageA\n"
            "# - packageB\n"
            "#\n"
            "export KEY=\"VALUE\"\n"
            "App -x $@\n"
        )
    )
], ids=[
    "minimal",
    "with-path",
    "with-command",
    "with-packages",
    "with-command-and-packages",
], indirect=[
    "packages"
])
def test_export_bash_script(temporary_directory, options, packages, expected):
    """Export Bash script."""
    options.update({"packages": packages})

    wiz.export_script(
        temporary_directory, "bash", "foo", **options
    )

    file_path = os.path.join(temporary_directory, "foo")
    assert os.path.isfile(file_path) is True

    with open(file_path, "r") as stream:
        assert stream.read() == expected


def test_export_bash_script_environ_error(temporary_directory):
    """Fail to export Bash script with empty environment mapping."""
    with pytest.raises(ValueError) as error:
        wiz.export_script(temporary_directory, "bash", "foo", {})

    assert "The environment mapping should not be empty." in str(error)


def test_export_script_error(temporary_directory):
    """Fail to export script with incorrect type."""
    with pytest.raises(ValueError) as error:
        wiz.export_script(
            temporary_directory, "duh", "foo",
            {"KEY": "VALUE"}
        )

    assert "'duh' is not a valid script type." in str(error)



# :coding: utf-8

import os
import io
import gzip

import pytest

import wiz.filesystem


@pytest.fixture()
def mocked_ensure_directory(mocker):
    """Return mocked ensure_directory function."""
    return mocker.patch.object(wiz.filesystem, "ensure_directory")


@pytest.fixture()
def mocked_os_access(mocker):
    """Return mocked os.access function."""
    return mocker.patch.object(os, "access")


@pytest.fixture()
def mocked_io_open(mocker):
    """Return mocked io.open function."""
    mocked_function = mocker.patch.object(io, "open")
    stream = mocker.MagicMock()
    mocked_function.return_value.__enter__.return_value = stream
    return {
        "func": mocked_function,
        "stream": stream
    }


@pytest.fixture()
def mocked_gzip_open(mocker):
    """Return mocked gzip.open function."""
    mocked_function = mocker.patch.object(gzip, "open")
    stream = mocker.MagicMock()
    mocked_function.return_value.__enter__.return_value = stream
    return {
        "func": mocked_function,
        "stream": stream
    }


@pytest.mark.parametrize("path, resolved_path, directory", [
    ("/path/to/file", "/path/to/file", "/path/to"),
    ("/path/to/somewhere/../else/file", "/path/to/else/file", "/path/to/else"),
    ("~/file", "/usr/people/me/file", "/usr/people/me"),
], ids=[
    "simple",
    "relative",
    "with-user-shortcut"
])
def test_export(
    path, resolved_path, directory, mocked_io_open, mocked_gzip_open,
    mocked_ensure_directory, monkeypatch
):
    """Export a file with content."""
    monkeypatch.setenv("HOME", "/usr/people/me")

    content = "THIS IS\n A TEST.\n"

    wiz.filesystem.export(path, content)

    mocked_ensure_directory.assert_called_once_with(directory)
    mocked_io_open["func"].assert_called_once_with(
        resolved_path, "w", encoding="utf8"
    )
    mocked_gzip_open["func"].assert_not_called()
    mocked_io_open["stream"].write.assert_called_once_with(content)


@pytest.mark.parametrize("path, resolved_path, directory", [
    ("/path/to/file", "/path/to/file", "/path/to"),
    ("/path/to/somewhere/../else/file", "/path/to/else/file", "/path/to/else"),
    ("~/file", "/usr/people/me/file", "/usr/people/me"),
], ids=[
    "simple",
    "relative",
    "with-user-shortcut"
])
def test_export_with_compression(
    path, resolved_path, directory, mocked_io_open, mocked_gzip_open,
    mocked_ensure_directory, monkeypatch
):
    """Export a compressed file with content."""
    monkeypatch.setenv("HOME", "/usr/people/me")

    content = "THIS IS\n A TEST.\n"

    wiz.filesystem.export(path, content, compressed=True)

    mocked_ensure_directory.assert_called_once_with(directory)
    mocked_gzip_open["func"].assert_called_once_with(resolved_path, "wb")
    mocked_io_open["func"].assert_not_called()
    mocked_gzip_open["stream"].write.assert_called_once_with(content)


def test_accessible(temporary_directory, temporary_file, mocked_os_access):
    """Indicate whether directory is accessible."""
    mocked_os_access.return_value = True
    assert wiz.filesystem.is_accessible(temporary_directory) is True
    assert wiz.filesystem.is_accessible(temporary_file) is False
    assert wiz.filesystem.is_accessible("/incorrect") is False

    mocked_os_access.return_value = False
    assert wiz.filesystem.is_accessible(temporary_directory) is False


def test_ensure_directory(temporary_directory, temporary_file):
    """Create directory if it doesn't exist."""
    path1 = os.path.join(temporary_directory, "path1")
    path2 = os.path.join(path1, "path2")
    assert os.path.isdir(path1) is False
    assert os.path.isdir(path2) is False

    wiz.filesystem.ensure_directory(path2)
    assert os.path.isdir(path1) is True
    assert os.path.isdir(path2) is True

    # Do not raise when folder exists
    wiz.filesystem.ensure_directory(path2)

    # Raise when element is a file
    with pytest.raises(OSError):
        wiz.filesystem.ensure_directory(temporary_file)

    # Raise for other errors
    with pytest.raises(OSError):
        wiz.filesystem.ensure_directory("/incorrect")


def test_sanitize_value():
    """Sanitize value."""
    value = "/path/to/a-file/with: A F@#%ing Name!!!"
    assert wiz.filesystem.sanitize_value(value) == (
        "/path/to/a-file/with:_A_F__%ing_Name___"
    )

    assert wiz.filesystem.sanitize_value(value, substitution_character="-") == (
        "/path/to/a-file/with:-A-F--%ing-Name---"
    )

    assert wiz.filesystem.sanitize_value(value, case_sensitive=False) == (
        "/path/to/a-file/with:_a_f__%ing_name___"
    )



# :coding: utf-8

import os

import wiz.config
import wiz.exception
import wiz.filesystem
import wiz.logging
import wiz.utility


def get_local():
    """Return the local registry if available.

    :return: :file:`~/.wiz/registry` or None.

    """
    registry_path = os.path.join(os.path.expanduser("~"), ".wiz", "registry")
    if os.path.isdir(registry_path) and os.access(registry_path, os.R_OK):
        return registry_path


def get_defaults():
    """Return the default registries.

    :return: List of default registry paths.

    .. seealso:: :ref:`configuration/registry_paths`

    """
    config = wiz.config.fetch()
    return config.get("registry", {}).get("paths", [])


def fetch(paths, include_local=True, include_working_directory=True):
    """Fetch all registries from *paths*.

    :param paths: List of paths to consider as registry paths if available.

    :param include_local: Indicate whether the local registry should be
        included.

    :param include_working_directory: Indicate whether the current working
        directory should be parsed to discover registry folders.

    :return: List of valid registry paths.

    """
    registries = []

    for path in paths:
        path = os.path.abspath(path)

        if not wiz.filesystem.is_accessible(path):
            continue
        registries.append(path)

    if include_working_directory:
        for registry_path in discover(os.getcwd()):
            registries.append(registry_path)

    registry_path = get_local()
    if registry_path and include_local:
        registries.append(registry_path)

    return registries


def discover(path):
    """Yield available registry folders from *path* folder hierarchy.

    Each folder constituting the hierarchy of *path* are parsed so that
    existing :file:`.wiz/registry` folders can be yield from the deepest
    to the closest.

    Example::

        >>> list(discover("/jobs/ads/project/identity/shot"))
        [
            "/jobs/ads/project/.wiz/registry",
            "/jobs/ads/project/identity/shot/.wiz/registry"
        ]

    :param path: Path to discover registries from.

    :return: List of valid registry paths.

    .. seealso:: :ref:`registry/discover-implicit`

    """
    path = os.path.abspath(path)

    config = wiz.config.fetch()

    # Only discover the registry if the top level hierarchy is defined in the
    # config as 'discovery_prefix'.
    prefix = config.get("registry", {}).get("discovery_prefix", os.sep)
    if not path.startswith(prefix):
        return

    _path = [p for p in path[len(prefix):].split(os.sep) if len(p)]
    for folder in _path:
        prefix = os.path.join(prefix, folder)
        registry_path = os.path.join(prefix, ".wiz", "registry")

        if wiz.filesystem.is_accessible(registry_path):
            yield registry_path


def install_to_path(definitions, registry_path, overwrite=False):
    """Install a list of definitions to a registry on the file system.

    :param definitions: List of :class:`wiz.definition.Definition` instances.

    :param registry_path: Targeted registry path to install to

    :param overwrite: Indicate whether existing definitions in the target
        registry should be overwritten. Default is False.

    :raise: :exc:`wiz.exception.DefinitionsExist` if definitions already exist
        in the target registry and overwrite is False.

    :raise: :exc:`wiz.exception.InstallNoChanges` if definitions already exists
        in the target registry and no changes is detected.

    :raise: :exc:`wiz.exception.InstallError` if the target registry path is not
        a valid directory.

    :raise: :exc:`OSError` if definitions can not be exported in
        *registry_path*.

    """
    logger = wiz.logging.Logger(__name__ + ".install_to_path")

    if not os.path.isdir(registry_path):
        raise wiz.exception.InstallError(
            "{!r} is not a valid registry directory.".format(registry_path)
        )

    registry_path = os.path.abspath(registry_path)

    # Record definitions to install.
    _definitions = []

    # Record existing definitions per identifier.
    existing = {}

    # Fetch all definitions from registry.
    mapping = wiz.fetch_definition_mapping([registry_path])

    for definition in definitions:
        request = definition.qualified_version_identifier

        try:
            _definition = wiz.fetch_definition(request, mapping)

        except wiz.exception.RequestNotFound:
            _definitions.append(definition)

        else:
            data1 = definition.data(copy_data=False)
            data2 = _definition.data(copy_data=False)
            if data1 == data2:
                continue

            existing[definition.qualified_identifier] = _definition
            _definitions.append(definition)

    # If no content was released.
    if len(_definitions) == 0:
        raise wiz.exception.InstallNoChanges()

    # Fail if overwrite is False and some definition paths exist in registry.
    if len(existing) > 0 and overwrite is False:
        raise wiz.exception.DefinitionsExist([
            wiz.utility.compute_label(definition)
            for definition in existing.values()
        ])

    # Release definitions
    for definition in _definitions:
        path = registry_path
        identifier = definition.identifier

        # Replace existing definition if necessary.
        if identifier in existing.keys():
            existing_definition_path = existing[identifier].path
            path = os.path.dirname(existing_definition_path)
            os.remove(existing_definition_path)

        wiz.export_definition(path, definition, overwrite=True)

    logger.info(
        "Successfully installed {number} definition(s) to "
        "registry {registry!r}.".format(
            number=len(_definitions),
            registry=registry_path
        )
    )



# :coding: utf-8

import pytest

import wiz.definition
import wiz.environ
import wiz.exception
import wiz.package
from wiz.utility import Requirement, Version


@pytest.fixture()
def mocked_definition_query(mocker):
    """Return mocked definition getter."""
    return mocker.patch.object(wiz.definition, "query")


@pytest.fixture()
def mocked_package_constructor(mocker):
    """Return mocked Package constructor."""
    return mocker.patch.object(wiz.package, "Package")


@pytest.fixture()
def mocked_package_create(mocker):
    """Return mocked :func:`wiz.package.create`."""
    return mocker.patch.object(wiz.package, "create")


@pytest.fixture()
def mocked_combine_environ(mocker):
    """Return mocked combine_environ_mapping function."""
    return mocker.patch.object(
        wiz.package, "combine_environ_mapping", return_value={"KEY": "VALUE"}
    )


@pytest.fixture()
def mocked_combine_command(mocker):
    """Return mocked combine_command_mapping function."""
    return mocker.patch.object(
        wiz.package, "combine_command_mapping", return_value={"APP": "APP_EXE"}
    )


@pytest.fixture()
def mocked_environ_sanitize(mocker):
    """Return mocked 'wiz.environ.sanitize' function."""
    return mocker.patch.object(
        wiz.environ, "sanitize", return_value={"CLEAN_KEY": "CLEAN_VALUE"}
    )


@pytest.mark.parametrize("kwargs, namespace_counter", [
    ({}, None),
    ({"namespace_counter": "COUNTER"}, "COUNTER")
], ids=[
    "simple",
    "with-namespace-hints",
])
def test_extract_without_variant(
    mocked_definition_query, mocked_package_create, kwargs, namespace_counter
):
    """Extract one Package from definition."""
    definition = wiz.definition.Definition({"identifier": "test",})
    mocked_definition_query.return_value = definition

    requirement = Requirement("test")
    packages = wiz.package.extract(
        requirement, "__DEFINITION_MAPPING__", **kwargs
    )

    mocked_definition_query.assert_called_once_with(
        requirement, "__DEFINITION_MAPPING__",
        namespace_counter=namespace_counter
    )

    mocked_package_create.assert_called_once_with(definition)
    assert packages == [mocked_package_create.return_value]


@pytest.mark.parametrize("kwargs, namespace_counter", [
    ({}, None),
    ({"namespace_counter": "COUNTER"}, "COUNTER")
], ids=[
    "simple",
    "with-namespace-hints",
])
def test_extract_with_all_variants(
    mocked_definition_query, mocked_package_create, kwargs, namespace_counter
):
    """Extract all variant Packages from definition."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "variants": [
            {"identifier": "V3"},
            {"identifier": "V2"},
            {"identifier": "V1"},
        ]
    })
    mocked_definition_query.return_value = definition

    requirement = Requirement("test")
    packages = wiz.package.extract(
        requirement, "__DEFINITION_MAPPING__", **kwargs
    )

    mocked_definition_query.assert_called_once_with(
        requirement, "__DEFINITION_MAPPING__",
        namespace_counter=namespace_counter
    )

    assert mocked_package_create.call_count == 3
    mocked_package_create.assert_any_call(definition, variant_identifier="V3")
    mocked_package_create.assert_any_call(definition, variant_identifier="V2")
    mocked_package_create.assert_any_call(definition, variant_identifier="V1")

    assert packages == [
        mocked_package_create.return_value,
        mocked_package_create.return_value,
        mocked_package_create.return_value,
    ]


@pytest.mark.parametrize("kwargs, namespace_counter", [
    ({}, None),
    ({"namespace_counter": "COUNTER"}, "COUNTER")
], ids=[
    "simple",
    "with-namespace-hints",
])
def test_extract_with_one_requested_variant(
    mocked_definition_query, mocked_package_create, kwargs, namespace_counter
):
    """Extract one requested variant Package from definition."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "variants": [
            {"identifier": "V3"},
            {"identifier": "V2"},
            {"identifier": "V1"},
        ]
    })
    mocked_definition_query.return_value = definition

    requirement = Requirement("test[V2]")
    packages = wiz.package.extract(
        requirement, "__DEFINITION_MAPPING__", **kwargs
    )

    mocked_definition_query.assert_called_once_with(
        requirement, "__DEFINITION_MAPPING__",
        namespace_counter=namespace_counter
    )

    mocked_package_create.assert_called_once_with(
        definition, variant_identifier="V2"
    )
    assert packages == [mocked_package_create.return_value]


def test_extract_context_without_packages(
    mocked_combine_environ, mocked_combine_command, mocked_environ_sanitize
):
    """Extract context with no packages."""
    assert wiz.package.extract_context([]) == {
        "environ": {"CLEAN_KEY": "CLEAN_VALUE"}
    }
    mocked_combine_environ.assert_not_called()
    mocked_combine_command.assert_not_called()
    mocked_environ_sanitize.assert_called_once_with({})


def test_extract_context_with_empty_package(
    mocked_combine_environ, mocked_combine_command, mocked_environ_sanitize
):
    """Extract context with package without environ nor command."""
    definition = wiz.definition.Definition({"identifier": "test"})
    packages = [wiz.package.Package(definition)]

    assert wiz.package.extract_context(packages) == {
        "command": {"APP": "APP_EXE"}, "environ": {"CLEAN_KEY": "CLEAN_VALUE"}
    }
    mocked_combine_environ.assert_called_once_with("test", {}, {})
    mocked_combine_command.assert_called_once_with("test", {}, {})
    mocked_environ_sanitize.assert_called_once_with({"KEY": "VALUE"})


def test_extract_context_with_one_package(
    mocked_combine_environ, mocked_combine_command, mocked_environ_sanitize
):
    """Extract context with one package."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "environ": {"key1": "value1"},
        "command": {"app": "App"}
    })
    packages = [wiz.package.Package(definition)]

    assert wiz.package.extract_context(packages) == {
        "command": {"APP": "APP_EXE"}, "environ": {"CLEAN_KEY": "CLEAN_VALUE"}
    }
    mocked_combine_environ.assert_called_once_with(
        "test", {}, {"key1": "value1"}
    )
    mocked_combine_command.assert_called_once_with(
        "test", {}, {"app": "App"}
    )
    mocked_environ_sanitize.assert_called_once_with({"KEY": "VALUE"})


def test_extract_context_with_six_package(
    mocked_combine_environ, mocked_combine_command, mocked_environ_sanitize
):
    """Extract context with five packages."""
    definitions = [
        wiz.definition.Definition({
            "identifier": "test1",
            "version": "0.1.0",
            "environ": {"key1": "value1"}}
        ),
        wiz.definition.Definition({
            "identifier": "test2",
            "command": {"app1": "App1"},
            "environ": {"key2": "value2"}
        }),
        wiz.definition.Definition({
            "identifier": "test3",
            "version": "3.1.2",
            "command": {"app2": "App2"},
            "environ": {"key3": "value3"}
        }),
        wiz.definition.Definition({
            "identifier": "test4",
            "version": "0.11.2",
            "environ": {"key4": "value4"}
        }),
        wiz.definition.Definition({
            "identifier": "test5",
            "version": "30",
            "install-location": "/path/to/package",
            "command": {"app1": "AppX"},
            "environ": {"PATH": "${INSTALL_LOCATION}/bin"}
        }),
        wiz.definition.Definition({
            "identifier": "test6",
            "version": "30.5",
            "install-location": "/path/to/package",
            "command": {"app1": "AppX"},
            "environ": {"PATH": "${INSTALL_LOCATION}/bin"}
        })
    ]

    packages = [wiz.package.create(definition) for definition in definitions]

    assert wiz.package.extract_context(packages) == {
        "command": {"APP": "APP_EXE"}, "environ": {"CLEAN_KEY": "CLEAN_VALUE"}
    }

    assert mocked_combine_environ.call_count == 6
    mocked_combine_environ.assert_any_call(
        "test1==0.1.0", {}, {"key1": "value1"}
    )
    mocked_combine_environ.assert_any_call(
        "test2", {"KEY": "VALUE"}, {"key2": "value2"}
    )
    mocked_combine_environ.assert_any_call(
        "test3==3.1.2", {"KEY": "VALUE"}, {"key3": "value3"}
    )
    mocked_combine_environ.assert_any_call(
        "test4==0.11.2", {"KEY": "VALUE"}, {"key4": "value4"}
    )
    mocked_combine_environ.assert_any_call(
        "test5==30", {"KEY": "VALUE"}, {"PATH": "/path/to/package/bin"}
    )
    mocked_combine_environ.assert_any_call(
        "test6==30.5", {"KEY": "VALUE"}, {"PATH": "/path/to/package/bin"}
    )

    assert mocked_combine_command.call_count == 6
    mocked_combine_command.assert_any_call(
        "test1==0.1.0", {}, {}
    )
    mocked_combine_command.assert_any_call(
        "test2", {"APP": "APP_EXE"}, {"app1": "App1"}
    )
    mocked_combine_command.assert_any_call(
        "test3==3.1.2", {"APP": "APP_EXE"}, {"app2": "App2"}
    )
    mocked_combine_command.assert_any_call(
        "test4==0.11.2", {"APP": "APP_EXE"}, {}
    )
    mocked_combine_command.assert_any_call(
        "test5==30", {"APP": "APP_EXE"}, {"app1": "AppX"}
    )
    mocked_combine_command.assert_any_call(
        "test6==30.5", {"APP": "APP_EXE"}, {"app1": "AppX"}
    )

    mocked_environ_sanitize.assert_called_once_with({"KEY": "VALUE"})


@pytest.mark.usefixtures("mocked_combine_command")
@pytest.mark.usefixtures("mocked_environ_sanitize")
def test_extract_context_with_initial_data(mocked_combine_environ):
    """Return extracted context with initial environ mapping."""
    definitions = [
        wiz.definition.Definition({
            "identifier": "test1",
            "version": "0.1.0",
            "environ": {"key1": "value1"}}
        ),
        wiz.definition.Definition({
            "identifier": "test2",
            "command": {"app1": "App1"},
            "environ": {"key2": "value2"}
        }),
        wiz.definition.Definition({
            "identifier": "test3",
            "version": "3.1.2",
            "command": {"app2": "App2"},
            "environ": {"key3": "value3"}
        }),
    ]

    packages = [wiz.package.create(definition) for definition in definitions]

    assert wiz.package.extract_context(
        packages, environ_mapping={"INITIAL_KEY": "INITIAL_VALUE"}
    ) == {
        "command": {"APP": "APP_EXE"},
        "environ": {"CLEAN_KEY": "CLEAN_VALUE"}
    }

    mocked_combine_environ.assert_any_call(
        "test1==0.1.0", {"INITIAL_KEY": "INITIAL_VALUE"}, {"key1": "value1"}
    )


@pytest.mark.parametrize("mapping1, mapping2, expected, warning", [
    (
        {"KEY": "HELLO"},
        {"KEY": "${KEY} WORLD!"},
        {"KEY": "HELLO WORLD!"},
        None
    ),
    (
        {"KEY": "HELLO"},
        {"KEY": "$KEY WORLD!"},
        {"KEY": "HELLO WORLD!"},
        None
    ),
    (
        {"KEY": "VALUE1"},
        {"KEY": "VALUE2"},
        {"KEY": "VALUE2"},
        "The 'KEY' variable is being overridden in '{package}'"
    ),
    (
        {"PLUGIN": "/path/to/settings", "HOME": "/usr/people/me"},
        {"PLUGIN": "${HOME}/.app:${PLUGIN}"},
        {
            "HOME": "/usr/people/me",
            "PLUGIN": "/usr/people/me/.app:/path/to/settings"
        },
        None
    ),
    (
        {"PLUGIN": "/path/to/settings", "HOME": "/usr/people/me"},
        {"PLUGIN": "$HOME/.app:$PLUGIN"},
        {
            "HOME": "/usr/people/me",
            "PLUGIN": "/usr/people/me/.app:/path/to/settings"
        },
        None
    )
], ids=[
    "combine-key",
    "combine-key-without-curly-brackets",
    "override-key",
    "mixed-combination",
    "mixed-combination-without-curly-brackets"
])
def test_combine_environ_mapping(logger, mapping1, mapping2, expected, warning):
    """Return combined environment mapping from *mapping1* and *mapping2*."""
    package_identifier = "Test==0.1.0"

    result = wiz.package.combine_environ_mapping(
        package_identifier, mapping1, mapping2
    )
    assert result == expected

    if warning is None:
        logger.warning.assert_not_called()
    else:
        logger.warning.assert_called_once_with(
            warning.format(package=package_identifier)
        )


@pytest.mark.parametrize("mapping1, mapping2, expected", [
    (
        {"app1": "App1"},
        {"app2": "App2"},
        {"app1": "App1", "app2": "App2"},
    ),
    (
        {"app1": "App1.0"},
        {"app1": "App1.5"},
        {"app1": "App1.5"},
    )
], ids=[
    "combine-key",
    "override-key",
])
def test_combine_command_mapping(mapping1, mapping2, expected):
    """Return combined command mapping from *mapping1* and *mapping2*."""
    package_identifier = "Test==0.1.0"
    assert wiz.package.combine_command_mapping(
        package_identifier, mapping1, mapping2
    ) == expected


def test_create_package(mocked_package_constructor):
    """Create package."""
    definition = wiz.definition.Definition({"identifier": "test"})

    package = wiz.package.create(definition)
    assert package == mocked_package_constructor.return_value
    mocked_package_constructor.assert_called_once_with(definition)


def test_create_package_with_variant(mocked_package_constructor):
    """Create package with variant."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "variants": [
            {"identifier": "V2"},
            {"identifier": "V1"}
        ]
    })

    package = wiz.package.create(definition, variant_identifier="V1")
    assert package == mocked_package_constructor.return_value
    mocked_package_constructor.assert_called_once_with(
        definition, variant_index=1
    )


def test_create_package_with_variant_error():
    """Fail to create package with variant."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "version": "0.1.0",
        "namespace": "foo",
        "variants": [
            {"identifier": "V2"},
            {"identifier": "V1"}
        ]
    })

    with pytest.raises(wiz.exception.RequestNotFound) as error:
        wiz.package.create(definition, variant_identifier="invalid")

    assert (
        "The variant 'invalid' could not been resolved for 'foo::test==0.1.0'"
    ) in str(error)


def test_incorrect_package():
    """Fail to create a package."""
    # Requesting non exiting variant
    definition = wiz.definition.Definition({"identifier": "test"})

    with pytest.raises(wiz.exception.PackageError) as error:
        wiz.package.Package(definition, variant_index=3)

    assert (
        "Package cannot be created from definition 'test' with variant "
        "index #3."
    ) in str(error)

    # Failing to request variant
    definition = wiz.definition.Definition({
        "identifier": "test",
        "variants": [{"identifier": "V1"}]
    })

    with pytest.raises(wiz.exception.PackageError) as error:
        wiz.package.Package(definition)

    assert (
        "Package cannot be created from definition 'test' as no variant "
        "index is defined."
    ) in str(error)


def test_minimal_package():
    """Create a minimal package."""
    definition = wiz.definition.Definition({"identifier": "test"})

    package = wiz.package.Package(definition)
    assert package.definition == definition
    assert package.identifier == "test"
    assert package.variant is None
    assert package.variant_identifier is None
    assert package.version is None
    assert package.description is None
    assert package.namespace is None
    assert package.install_location is None
    assert package.environ == {}
    assert package.command == {}
    assert package.requirements == []
    assert package.conditions == []
    assert package.conditions_processed is False

    assert package.data() == {"identifier": "test"}
    assert package.localized_environ() == {}


def test_minimal_variant_package():
    """Create a minimal variant package."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "variants": [{"identifier": "V1"}]
    })

    package = wiz.package.Package(definition, variant_index=0)
    assert package.definition == definition
    assert package.identifier == "test[V1]"
    assert package.variant == definition.variants[0]
    assert package.variant_identifier == "V1"
    assert package.version is None
    assert package.description is None
    assert package.namespace is None
    assert package.install_location is None
    assert package.environ == {}
    assert package.command == {}
    assert package.requirements == []
    assert package.conditions == []
    assert package.conditions_processed is False

    assert package.data() == {
        "identifier": "test[V1]",
        "variant-identifier": "V1"
    }
    assert package.localized_environ() == {}


def test_minimal_package_with_namespace():
    """Create a minimal package with namespace."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "namespace": "foo",
    })

    package = wiz.package.Package(definition)
    assert package.definition == definition
    assert package.identifier == "foo::test"
    assert package.variant is None
    assert package.variant_identifier is None
    assert package.version is None
    assert package.description is None
    assert package.namespace == "foo"
    assert package.install_location is None
    assert package.environ == {}
    assert package.command == {}
    assert package.requirements == []
    assert package.conditions == []
    assert package.conditions_processed is False

    assert package.data() == {
        "identifier": "foo::test",
        "namespace": "foo",
    }
    assert package.localized_environ() == {}


def test_minimal_variant_package_with_namespace():
    """Create a minimal variant package with namespace."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "namespace": "foo",
        "variants": [{"identifier": "V1"}]
    })

    package = wiz.package.Package(definition, variant_index=0)
    assert package.definition == definition
    assert package.identifier == "foo::test[V1]"
    assert package.variant == definition.variants[0]
    assert package.variant_identifier == "V1"
    assert package.version is None
    assert package.description is None
    assert package.namespace == "foo"
    assert package.install_location is None
    assert package.environ == {}
    assert package.command == {}
    assert package.requirements == []
    assert package.conditions == []
    assert package.conditions_processed is False

    assert package.data() == {
        "identifier": "foo::test[V1]",
        "namespace": "foo",
        "variant-identifier": "V1"
    }
    assert package.localized_environ() == {}


def test_package_with_version():
    """Create a package with version."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "version": "0.1.0",
    })

    package = wiz.package.Package(definition)
    assert package.definition == definition
    assert package.identifier == "test==0.1.0"
    assert package.variant is None
    assert package.variant_identifier is None
    assert package.version == Version("0.1.0")
    assert package.description is None
    assert package.namespace is None
    assert package.install_location is None
    assert package.environ == {}
    assert package.command == {}
    assert package.requirements == []
    assert package.conditions == []
    assert package.conditions_processed is False

    assert package.data() == {
        "identifier": "test==0.1.0",
        "version": "0.1.0",
    }
    assert package.localized_environ() == {}


def test_variant_package_with_version():
    """Create a variant package with version."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "version": "0.1.0",
        "variants": [{"identifier": "V1"}]
    })

    package = wiz.package.Package(definition, variant_index=0)
    assert package.definition == definition
    assert package.identifier == "test[V1]==0.1.0"
    assert package.variant == definition.variants[0]
    assert package.variant_identifier == "V1"
    assert package.version == Version("0.1.0")
    assert package.description is None
    assert package.namespace is None
    assert package.install_location is None
    assert package.environ == {}
    assert package.command == {}
    assert package.requirements == []
    assert package.conditions == []
    assert package.conditions_processed is False

    assert package.data() == {
        "identifier": "test[V1]==0.1.0",
        "version": "0.1.0",
        "variant-identifier": "V1"
    }
    assert package.localized_environ() == {}


def test_package_with_version_and_namespace():
    """Create a package with version and namespace."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "namespace": "foo",
        "version": "0.1.0",
    })

    package = wiz.package.Package(definition)
    assert package.definition == definition
    assert package.identifier == "foo::test==0.1.0"
    assert package.variant is None
    assert package.variant_identifier is None
    assert package.version == Version("0.1.0")
    assert package.description is None
    assert package.namespace == "foo"
    assert package.install_location is None
    assert package.environ == {}
    assert package.command == {}
    assert package.requirements == []
    assert package.conditions == []
    assert package.conditions_processed is False

    assert package.data() == {
        "identifier": "foo::test==0.1.0",
        "namespace": "foo",
        "version": "0.1.0",
    }
    assert package.localized_environ() == {}


def test_variant_package_with_version_and_namespace():
    """Create a variant package with version and namespace."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "version": "0.1.0",
        "namespace": "foo",
        "variants": [{"identifier": "V1"}]
    })

    package = wiz.package.Package(definition, variant_index=0)
    assert package.definition == definition
    assert package.identifier == "foo::test[V1]==0.1.0"
    assert package.variant == definition.variants[0]
    assert package.variant_identifier == "V1"
    assert package.version == Version("0.1.0")
    assert package.description is None
    assert package.namespace == "foo"
    assert package.install_location is None
    assert package.environ == {}
    assert package.command == {}
    assert package.requirements == []
    assert package.conditions == []
    assert package.conditions_processed is False

    assert package.data() == {
        "identifier": "foo::test[V1]==0.1.0",
        "version": "0.1.0",
        "namespace": "foo",
        "variant-identifier": "V1"
    }
    assert package.localized_environ() == {}


def test_package_with_description():
    """Create a package with description."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "description": "This is a definition"
    })

    package = wiz.package.Package(definition)
    assert package.definition == definition
    assert package.identifier == "test"
    assert package.variant is None
    assert package.variant_identifier is None
    assert package.version is None
    assert package.description == "This is a definition"
    assert package.namespace is None
    assert package.install_location is None
    assert package.environ == {}
    assert package.command == {}
    assert package.requirements == []
    assert package.conditions == []
    assert package.conditions_processed is False

    assert package.data() == {
        "identifier": "test",
        "description": "This is a definition",
    }
    assert package.localized_environ() == {}


def test_variant_package_with_description():
    """Create a variant package with description."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "description": "This is a definition",
        "variants": [{"identifier": "V1"}]
    })

    package = wiz.package.Package(definition, variant_index=0)
    assert package.definition == definition
    assert package.identifier == "test[V1]"
    assert package.variant == definition.variants[0]
    assert package.variant_identifier == "V1"
    assert package.version is None
    assert package.description == "This is a definition"
    assert package.namespace is None
    assert package.install_location is None
    assert package.environ == {}
    assert package.command == {}
    assert package.requirements == []
    assert package.conditions == []
    assert package.conditions_processed is False

    assert package.data() == {
        "identifier": "test[V1]",
        "description": "This is a definition",
        "variant-identifier": "V1"
    }
    assert package.localized_environ() == {}


def test_package_with_install_location():
    """Create a package with installation location."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "install-location": "/path/to/install"
    })

    package = wiz.package.Package(definition)
    assert package.definition == definition
    assert package.identifier == "test"
    assert package.variant is None
    assert package.variant_identifier is None
    assert package.version is None
    assert package.description is None
    assert package.namespace is None
    assert package.install_location == "/path/to/install"
    assert package.environ == {}
    assert package.command == {}
    assert package.requirements == []
    assert package.conditions == []
    assert package.conditions_processed is False

    assert package.data() == {
        "identifier": "test",
        "install-location": "/path/to/install"
    }
    assert package.localized_environ() == {}


def test_variant_package_with_install_location():
    """Create a variant package with installation location."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "install-location": "/path/to/install",
        "variants": [{
            "identifier": "V1"
        }]
    })

    package = wiz.package.Package(definition, variant_index=0)
    assert package.definition == definition
    assert package.identifier == "test[V1]"
    assert package.variant == definition.variants[0]
    assert package.variant_identifier == "V1"
    assert package.version is None
    assert package.description is None
    assert package.namespace is None
    assert package.install_location == "/path/to/install"
    assert package.environ == {}
    assert package.command == {}
    assert package.requirements == []
    assert package.conditions == []
    assert package.conditions_processed is False

    assert package.data() == {
        "identifier": "test[V1]",
        "install-location": "/path/to/install",
        "variant-identifier": "V1"
    }
    assert package.localized_environ() == {}


def test_variant_package_with_install_location_overwrite():
    """Create a variant package with installation location overwritten."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "install-location": "/path/to/install",
        "variants": [{
            "identifier": "V1",
            "install-location": "/path/to/installV1",
        }]
    })

    package = wiz.package.Package(definition, variant_index=0)
    assert package.definition == definition
    assert package.identifier == "test[V1]"
    assert package.variant == definition.variants[0]
    assert package.variant_identifier == "V1"
    assert package.version is None
    assert package.description is None
    assert package.namespace is None
    assert package.install_location == "/path/to/installV1"
    assert package.environ == {}
    assert package.command == {}
    assert package.requirements == []
    assert package.conditions == []
    assert package.conditions_processed is False

    assert package.data() == {
        "identifier": "test[V1]",
        "install-location": "/path/to/installV1",
        "variant-identifier": "V1"
    }
    assert package.localized_environ() == {}


def test_package_with_command():
    """Create a package with command."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "command": {
            "app": "App0.1",
        }
    })

    package = wiz.package.Package(definition)
    assert package.definition == definition
    assert package.identifier == "test"
    assert package.variant is None
    assert package.variant_identifier is None
    assert package.version is None
    assert package.description is None
    assert package.namespace is None
    assert package.install_location is None
    assert package.environ == {}
    assert package.command == {"app": "App0.1"}
    assert package.requirements == []
    assert package.conditions == []
    assert package.conditions_processed is False

    assert package.data() == {
        "identifier": "test",
        "command": {
            "app": "App0.1",
        }
    }
    assert package.localized_environ() == {}


def test_variant_package_with_command(mocked_combine_command):
    """Create a variant package with command."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "command": {
            "app": "App0.1",
        },
        "variants": [{
            "identifier": "V1",
            "command": {
                "appV1": "AppV10.1",
            }
        }]
    })

    package = wiz.package.Package(definition, variant_index=0)
    assert package.definition == definition
    assert package.identifier == "test[V1]"
    assert package.variant == definition.variants[0]
    assert package.variant_identifier == "V1"
    assert package.version is None
    assert package.description is None
    assert package.namespace is None
    assert package.install_location is None
    assert package.environ == {}
    assert package.command == mocked_combine_command.return_value
    assert package.requirements == []
    assert package.conditions == []
    assert package.conditions_processed is False

    assert package.data() == {
        "identifier": "test[V1]",
        "command": mocked_combine_command.return_value,
        "variant-identifier": "V1"
    }
    assert package.localized_environ() == {}


def test_package_with_environ():
    """Create a package with environment mapping."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "environ": {
            "KEY1": "VALUE1",
        }
    })

    package = wiz.package.Package(definition)
    assert package.definition == definition
    assert package.identifier == "test"
    assert package.variant is None
    assert package.variant_identifier is None
    assert package.version is None
    assert package.description is None
    assert package.namespace is None
    assert package.install_location is None
    assert package.environ == {"KEY1": "VALUE1"}
    assert package.command == {}
    assert package.requirements == []
    assert package.conditions == []
    assert package.conditions_processed is False

    assert package.data() == {
        "identifier": "test",
        "environ": {
            "KEY1": "VALUE1",
        }
    }
    assert package.localized_environ() == {"KEY1": "VALUE1"}


def test_variant_package_with_environ(mocked_combine_environ):
    """Create a variant package with environment mapping."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "environ": {
            "KEY1": "VALUE1",
        },
        "variants": [{
            "identifier": "V1",
            "environ": {
                "KEY_V1": "VALUE_V1",
            },
        }]
    })

    package = wiz.package.Package(definition, variant_index=0)
    assert package.definition == definition
    assert package.identifier == "test[V1]"
    assert package.variant == definition.variants[0]
    assert package.variant_identifier == "V1"
    assert package.version is None
    assert package.description is None
    assert package.namespace is None
    assert package.install_location is None
    assert package.environ == mocked_combine_environ.return_value
    assert package.command == {}
    assert package.requirements == []
    assert package.conditions == []
    assert package.conditions_processed is False

    assert package.data() == {
        "identifier": "test[V1]",
        "environ": mocked_combine_environ.return_value,
        "variant-identifier": "V1"
    }
    assert package.localized_environ() == mocked_combine_environ.return_value


def test_package_with_requirements():
    """Create a package with requirements."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "requirements": [
            "envA >= 1.0.0",
            "envB >= 3.4.2, < 4",
            "envC"
        ]
    })

    package = wiz.package.Package(definition)
    assert package.definition == definition
    assert package.identifier == "test"
    assert package.variant is None
    assert package.variant_identifier is None
    assert package.version is None
    assert package.description is None
    assert package.namespace is None
    assert package.install_location is None
    assert package.environ == {}
    assert package.command == {}
    assert package.requirements == [
        Requirement("envA >= 1.0.0"),
        Requirement("envB >= 3.4.2, < 4"),
        Requirement("envC")
    ]
    assert package.conditions == []
    assert package.conditions_processed is False

    assert package.data() == {
        "identifier": "test",
        "requirements": [
            "envA >= 1.0.0",
            "envB >= 3.4.2, < 4",
            "envC"
        ]
    }
    assert package.localized_environ() == {}


def test_variant_package_with_requirements():
    """Create a variant package with requirements."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "requirements": [
            "envA >= 1.0.0",
        ],
        "variants": [{
            "identifier": "V1",
            "requirements": [
                "envC"
            ]
        }]
    })

    package = wiz.package.Package(definition, variant_index=0)
    assert package.definition == definition
    assert package.identifier == "test[V1]"
    assert package.variant == definition.variants[0]
    assert package.variant_identifier == "V1"
    assert package.version is None
    assert package.description is None
    assert package.namespace is None
    assert package.install_location is None
    assert package.environ == {}
    assert package.command == {}
    assert package.requirements == [
        Requirement("envA >= 1.0.0"),
        Requirement("envC"),
    ]
    assert package.conditions == []
    assert package.conditions_processed is False

    assert package.data() == {
        "identifier": "test[V1]",
        "requirements": [
            "envA >= 1.0.0",
            "envC"
        ],
        "variant-identifier": "V1"
    }
    assert package.localized_environ() == {}


def test_package_with_conditions():
    """Create a package with conditions."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "conditions": [
            "envA >= 1.0.0",
            "envB >= 3.4.2, < 4",
            "envC"
        ]
    })

    package = wiz.package.Package(definition)
    assert package.definition == definition
    assert package.identifier == "test"
    assert package.variant is None
    assert package.variant_identifier is None
    assert package.version is None
    assert package.description is None
    assert package.namespace is None
    assert package.install_location is None
    assert package.environ == {}
    assert package.command == {}
    assert package.requirements == []
    assert package.conditions == [
        Requirement("envA >= 1.0.0"),
        Requirement("envB >= 3.4.2, < 4"),
        Requirement("envC")
    ]
    assert package.conditions_processed is False

    assert package.data() == {
        "identifier": "test",
        "conditions": [
            "envA >= 1.0.0",
            "envB >= 3.4.2, < 4",
            "envC"
        ]
    }
    assert package.localized_environ() == {}

    package.conditions_processed = True
    assert package.conditions_processed is True


def test_variant_package_with_conditions():
    """Create a variant package with conditions."""
    definition = wiz.definition.Definition({
        "identifier": "test",
        "conditions": [
            "envA >= 1.0.0",
        ],
        "variants": [{"identifier": "V1"}]
    })

    package = wiz.package.Package(definition, variant_index=0)
    assert package.definition == definition
    assert package.identifier == "test[V1]"
    assert package.variant == definition.variants[0]
    assert package.variant_identifier == "V1"
    assert package.version is None
    assert package.description is None
    assert package.namespace is None
    assert package.install_location is None
    assert package.environ == {}
    assert package.command == {}
    assert package.requirements == []
    assert package.conditions == [
        Requirement("envA >= 1.0.0"),
    ]
    assert package.conditions_processed is False

    assert package.data() == {
        "identifier": "test[V1]",
        "conditions": [
            "envA >= 1.0.0",
        ],
        "variant-identifier": "V1"
    }
    assert package.localized_environ() == {}

    package.conditions_processed = True
    assert package.conditions_processed is True


def test_package_localized_environ():
    """Return localized environment."""
    definition = wiz.definition.Definition({
        "identifier": "foo",
        "install-location": "/path/to/package",
        "environ": {
            "PATH": "${INSTALL_LOCATION}/bin:${PATH}",
            "PYTHONPATH": (
                "${INSTALL_LOCATION}/lib/python2.7/site-packages:${PYTHONPATH}"
            )
        }
    })

    package = wiz.package.Package(definition)

    assert package.localized_environ() == {
        "PATH": "/path/to/package/bin:${PATH}",
        "PYTHONPATH": (
            "/path/to/package/lib/python2.7/site-packages:${PYTHONPATH}"
        )
    }


def test_package_localized_environ_with_root():
    """Return localized environment."""
    definition = wiz.definition.Definition({
        "identifier": "foo",
        "install-root": "/path/to/root",
        "install-location": "${INSTALL_ROOT}/data",
        "environ": {
            "PATH": "${INSTALL_LOCATION}/bin:${PATH}",
            "PYTHONPATH": (
                "${INSTALL_LOCATION}/lib/python2.7/site-packages:${PYTHONPATH}"
            )
        }
    })

    package = wiz.package.Package(definition)

    assert package.localized_environ() == {
        "PATH": "/path/to/root/data/bin:${PATH}",
        "PYTHONPATH": (
            "/path/to/root/data/lib/python2.7/site-packages:${PYTHONPATH}"
        )
    }


def test_package_localized_environ_without_key():
    """Return localized environment with 'install-location' key."""
    definition = wiz.definition.Definition({
        "identifier": "foo",
        "environ": {
            "PATH": "${INSTALL_LOCATION}/bin:${PATH}",
            "PYTHONPATH": (
                "${INSTALL_LOCATION}/lib/python2.7/site-packages:${PYTHONPATH}"
            )
        }
    })

    package = wiz.package.Package(definition)

    assert package.localized_environ() == {
        "PATH": "${INSTALL_LOCATION}/bin:${PATH}",
        "PYTHONPATH": (
            "${INSTALL_LOCATION}/lib/python2.7/site-packages:${PYTHONPATH}"
        )
    }



# :coding: utf-8

import copy
import datetime
import os
import platform
import time
import traceback
import json

from wiz.utility import Requirement, Version
from ._version import __version__

#: Indicate whether the history should be recorded.
_IS_HISTORY_RECORDED = False

#: Indicate whether actions should only include 'identifier' keyword.
_MINIMAL_ACTIONS_REQUIRED = False

#: Mapping containing the entire context resolution history report.
_HISTORY = {
    "version": __version__,
    "user": None,
    "hostname": None,
    "timestamp": None,
    "timezone": None,
    "command": None,
    "actions": []
}


def get(serialized=False):
    """Return recorded history mapping.

    :param serialized: Indicate whether the returned history should be
        serialized as a :term:`JSON` string. Default is False.

    :return: History mapping - serialized or not - in the form of
        ::

            {
                "version": "3.0.0",
                "user": "john-doe",
                "hostname": "ws123",
                "timestamp": "2020-08-14T10:56:58.529201",
                "timezone": "PDT",
                "command": "wiz --record /tmp use foo --view",
                "actions": [
                    ...
                ]
            }

    """
    if serialized:
        return json.dumps(_HISTORY, default=_json_default).encode("utf-8")
    return _HISTORY


def start_recording(command=None, minimal_actions=False):
    """Start recording the execution history.

    This command will add information about the execution context to the history
    mapping (username, hostname, time, timezone) and activate the recording
    of actions via :func:`record_action`.

    :param command: Indicate the command line which is being executed. Default
        is None.

    :param minimal_actions: Indicate whether actions should only include the
        'identifier' keyword and discard all other elements passed to
        :func:`record_action`. If False, the execution time will be longer when
        history is being recorded. Default is False.

    """
    global _IS_HISTORY_RECORDED
    _IS_HISTORY_RECORDED = True

    global _MINIMAL_ACTIONS_REQUIRED
    _MINIMAL_ACTIONS_REQUIRED = minimal_actions

    global _HISTORY
    _HISTORY = {
        "version": __version__,
        "user": os.environ.get("USER"),
        "hostname": platform.node(),
        "timestamp": datetime.datetime.now().isoformat(),
        "timezone": time.tzname[1],
        "command": None,
        "actions": []
    }

    if command is not None:
        _HISTORY["command"] = command


def stop_recording():
    """Stop recording the history."""
    global _IS_HISTORY_RECORDED
    _IS_HISTORY_RECORDED = False


def record_action(identifier, **kwargs):
    """Add an action to the history.

    The action is identified by its unique *identifier* and every relevant
    arguments which will be serialized to provide an accurate snapshot of the
    execution context.

    :param identifier: Unique identifier of the action.

    .. warning::

        This operation will be discarded if the history is not being
        :func:`recorded <start_recording>`.

    """
    if not _IS_HISTORY_RECORDED:
        return

    action = {"identifier": identifier}

    if not _MINIMAL_ACTIONS_REQUIRED:
        action.update(**kwargs)

        if isinstance(action.get("error"), Exception):
            action["traceback"] = traceback.format_exc().splitlines()

    global _HISTORY
    _HISTORY["actions"].append(copy.deepcopy(action))


def _json_default(_object):
    """Override :meth:`json.JSONEncoder.default` to serialize all objects.

    Usage::

        >>> import json
        >>> json.dumps(_object, default=_json_default).encode("utf-8")

    """
    from wiz.definition import Definition
    from wiz.package import Package
    from wiz.graph import Graph, Node, StoredNode

    if isinstance(_object, Definition):
        data = _object.data()
        data["path"] = _object.path
        data["registry_path"] = _object.registry_path
        return data

    if isinstance(_object, (Graph, Graph, Node, StoredNode, Package)):
        return _object.data()

    elif isinstance(_object, (Requirement, Version)):
        return str(_object)

    elif isinstance(_object, Exception):
        return str(_object)

    elif isinstance(_object, set):
        return list(_object)

    raise TypeError("{} is not JSON serializable.".format(_object))



# :coding: utf-8

import collections
import imp
import os
import uuid

import toml

import wiz.logging
import wiz.utility

#: Global configuration mapping.
_CONFIG = None


def fetch(refresh=False):
    """Fetch configuration mapping.

    The configuration created is cached for future usage so that configuration
    previously fetched will be returned.

    :param refresh: Indicate whether the configuration should be re-created
        instead of using configuration previously created whenever possible.
        Default is False.

    :return: Configuration mapping.

    """
    logger = wiz.logging.Logger(__name__ + ".fetch")

    global _CONFIG

    if _CONFIG is not None and not refresh:
        return _CONFIG

    config = {}

    # Fetch all configurations paths.
    root = os.path.dirname(__file__)
    paths = [
        os.path.join(root, "package_data", "config.toml"),
        os.path.join(os.path.expanduser("~"), ".wiz", "config.toml")
    ]

    for file_path in paths:
        if not os.path.isfile(file_path):
            continue

        try:
            wiz.utility.deep_update(config, toml.load(file_path))
        except Exception as error:
            logger.warning(
                "Failed to load configuration from \"{0}\" [{1}]"
                .format(file_path, error)
            )

    # Extend configuration mapping with plugins
    for plugin in _discover_plugins():
        try:
            plugin.register(config)
        except Exception as error:
            logger.warning(
                "Failed to register plugin from \"{0}\" [{1}]"
                .format(plugin.__file__, error)
            )

    _CONFIG = config
    return _CONFIG


def _discover_plugins():
    """Discover and return plugins.

    :return: List of plugins discovered.

    """
    logger = wiz.logging.Logger(__name__ + "._discover_plugins")

    # Fetch all plugin paths.
    root = os.path.dirname(__file__)
    paths = [
        os.path.join(root, "package_data", "plugins"),
        os.path.join(os.path.expanduser("~"), ".wiz", "plugins")
    ]

    plugins = collections.OrderedDict()

    for dir_path in paths:
        if not os.path.isdir(dir_path):
            continue

        for file_path in os.listdir(dir_path):
            name, extension = os.path.splitext(file_path)
            if extension != ".py":
                continue

            module_path = os.path.join(dir_path, file_path)
            unique_name = uuid.uuid4().hex

            try:
                module = imp.load_source(unique_name, module_path)
                plugins[module.IDENTIFIER] = module

            except Exception as error:
                logger.warning(
                    "Failed to load plugin from \"{0}\" [{1}]"
                    .format(module_path, error)
                )
                continue

    return plugins.values()



# :coding: utf-8

import re

import six
from packaging.version import VERSION_PATTERN

import wiz.exception

_REGEXP_VERSION = re.compile(
    "^{}$".format(VERSION_PATTERN), re.VERBOSE | re.IGNORECASE
)


def validate_definition(data):
    """Validate *data* mapping used to create a definition.

    An error will be raised if the *data* mapping cannot be used to create an
    instance of :class:`wiz.definition.Definition`..

    :param data: Mapping to validate.

    :raise: :exc:`wiz.exception.IncorrectDefinition` if the *data* mapping
        is incorrect.

    """
    keywords = {
        "identifier", "version", "namespace", "description", "auto-use",
        "disabled", "install-root", "install-location", "system", "command",
        "environ", "requirements", "conditions", "variants"
    }

    try:
        validate_type(data, dict)
        validate_keywords(data, keywords)

        validate_identifier_keyword(data)
        validate_version_keyword(data)
        validate_namespace_keyword(data)
        validate_description_keyword(data)
        validate_auto_use_keyword(data)
        validate_disabled_keyword(data)
        validate_install_root_keyword(data)
        validate_install_location_keyword(data)
        validate_system_keyword(data)
        validate_command_keyword(data)
        validate_environ_keyword(data)
        validate_requirements_keyword(data)
        validate_conditions_keyword(data)
        validate_variants_keyword(data)

    except ValueError as error:
        raise wiz.exception.DefinitionError(str(error))


def validate_identifier_keyword(data, variant_index=None):
    """Validate 'identifier' keyword within *data* mapping.

    A correct *data* mapping should be in the form of::

        {
            "identifier": "foo"
        }

    :param data: Mapping to validate.

    :param variant_index: Index number of the variant mapping if applicable.
        Default is None.

    :raise: :exc:`ValueError` if the *data* mapping is incorrect.

    """
    data = data.get("identifier")

    name = "'identifier'"
    if variant_index is not None:
        name = "'variants/{}/identifier'".format(variant_index)

    validate_required(data, label=name)
    validate_type(data, six.string_types, label=name)


def validate_version_keyword(data):
    """Validate 'version' keyword within *data* mapping.

    A correct *data* mapping should be in the form of::

        {
            "version": "0.1.0"
        }

    :param data: Mapping to validate.

    :raise: :exc:`ValueError` if the *data* mapping is incorrect.

    """
    data = data.get("version")
    validate_type(data, six.string_types, label="'version'")

    if data is not None:
        match = _REGEXP_VERSION.match(data)
        if not match:
            raise ValueError("Invalid version: '{}'".format(data))


def validate_namespace_keyword(data):
    """Validate 'namespace' keyword within *data* mapping.

    A correct *data* mapping should be in the form of::

        {
            "namespace": "foo"
        }

    :param data: Mapping to validate.

    :raise: :exc:`ValueError` if the *data* mapping is incorrect.

    """
    data = data.get("namespace")
    validate_type(data, six.string_types, label="'namespace'")


def validate_description_keyword(data):
    """Validate 'description' keyword within *data* mapping.

    A correct *data* mapping should be in the form of::

        {
            "description": "This is a description"
        }

    :param data: Mapping to validate.

    :raise: :exc:`ValueError` if the *data* mapping is incorrect.

    """
    data = data.get("description")
    validate_type(data, six.string_types, label="'description'")


def validate_auto_use_keyword(data):
    """Validate 'auto-use' keyword within *data* mapping.

    A correct *data* mapping should be in the form of::

        {
            "auto-use": true
        }

    :param data: Mapping to validate.

    :raise: :exc:`ValueError` if the *data* mapping is incorrect.

    """
    data = data.get("auto-use")
    validate_type(data, bool, label="'auto-use'")


def validate_disabled_keyword(data):
    """Validate 'disabled' keyword within *data* mapping.

    A correct *data* mapping should be in the form of::

        {
            "disabled": true
        }

    :param data: Mapping to validate.

    :raise: :exc:`ValueError` if the *data* mapping is incorrect.

    """
    data = data.get("disabled")
    validate_type(data, bool, label="'disabled'")


def validate_install_root_keyword(data):
    """Validate 'install-root' keyword within *data* mapping.

    A correct *data* mapping should be in the form of::

        {
            "install-root": "/path/to/install/root"
        }

    :param data: Mapping to validate.

    :raise: :exc:`ValueError` if the *data* mapping is incorrect.

    """
    data = data.get("install-root")
    validate_type(data, six.string_types, label="'install-root'")


def validate_install_location_keyword(data, variant_index=None):
    """Validate 'install-location' keyword within *data* mapping.

    A correct *data* mapping should be in the form of::

        {
            "install-location": "/path/to/install/location"
        }

    :param data: Mapping to validate.

    :param variant_index: Index number of the variant mapping if applicable.
        Default is None.

    :raise: :exc:`ValueError` if the *data* mapping is incorrect.

    """
    data = data.get("install-location")

    name = "'install-location'"
    if variant_index is not None:
        name = "'variants/{}/install-location'".format(variant_index)

    validate_type(data, six.string_types, label=name)


def validate_system_keyword(data):
    """Validate 'system' keyword within *data* mapping.

    A correct *data* mapping should be in the form of::

        {
            "system": {
                "platform": "linux"
            }
        }

    :param data: Mapping to validate.

    :raise: :exc:`ValueError` if the *data* mapping is incorrect.

    """
    _data = data.get("system")
    validate_type(_data, dict, label="'system'")
    validate_not_empty(_data, label="'system'")

    _data = data.get("system", {})
    validate_keywords(_data, {"platform", "os", "arch"}, label="'system'")

    _data = data.get("system", {}).get("platform")
    validate_type(_data, six.string_types, label="'system/platform'")

    _data = data.get("system", {}).get("os")
    validate_type(_data, six.string_types, label="'system/os'")

    _data = data.get("system", {}).get("arch")
    validate_type(_data, six.string_types, label="'system/arch'")


def validate_command_keyword(data, variant_index=None):
    """Validate 'command' keyword within *data* mapping.

    A correct *data* mapping should be in the form of::

        {
            "command": {
                "foo": "FooExe"
            }
        }

    :param data: Mapping to validate.

    :param variant_index: Index number of the variant mapping if applicable.
        Default is None.

    :raise: :exc:`ValueError` if the *data* mapping is incorrect.

    """
    data = data.get("command")

    name = "'command'"
    if variant_index is not None:
        name = "'variants/{}/command'".format(variant_index)

    validate_type(data, dict, label=name)
    validate_not_empty(data, label=name)


def validate_environ_keyword(data, variant_index=None):
    """Validate 'environ' keyword within *data* mapping.

    A correct *data* mapping should be in the form of::

        {
            "environ": {
                "KEY": "VALUE"
            }
        }

    :param data: Mapping to validate.

    :param variant_index: Index number of the variant mapping if applicable.
        Default is None.

    :raise: :exc:`ValueError` if the *data* mapping is incorrect.

    """
    data = data.get("environ")

    name = "'environ'"
    if variant_index is not None:
        name = "'variants/{}/environ'".format(variant_index)

    validate_type(data, dict, label=name)
    validate_not_empty(data, label=name)


def validate_requirements_keyword(data, variant_index=None):
    """Validate 'requirements' keyword within *data* mapping.

    A correct *data* mapping should be in the form of::

        {
            "requirements": [
                "foo == 1.0.0",
                "bar >= 1, < 2"
            ]
        }

    :param data: Mapping to validate.

    :param variant_index: Index number of the variant mapping if applicable.
        Default is None.

    :raise: :exc:`ValueError` if the *data* mapping is incorrect.

    """
    data = data.get("requirements")

    name = "'requirements'"
    if variant_index is not None:
        name = "'variants/{}/requirements'".format(variant_index)

    validate_type(data, list, label=name)
    validate_not_empty(data, label=name)


def validate_conditions_keyword(data):
    """Validate 'conditions' keyword within *data* mapping.

    A correct *data* mapping should be in the form of::

        {
            "conditions": [
                "bim >= 3.0.0"
            ]
        }

    :param data: Mapping to validate.

    :raise: :exc:`ValueError` if the *data* mapping is incorrect.

    """
    data = data.get("conditions")
    validate_type(data, list, label="'conditions'")
    validate_not_empty(data, label="'conditions'")


def validate_variants_keyword(data):
    """Validate 'variants' keyword within *data* mapping.

    A correct *data* mapping should be in the form of::

        {
            "variants": [
                {
                    "identifier": "foo",
                    "install-location": "/path/to/install/location",
                    "requirements": [
                        "bar >= 1, < 2"
                    ]
                }
            ]
        }

    :param data: Mapping to validate.

    :raise: :exc:`ValueError` if the *data* mapping is incorrect.

    """
    data = data.get("variants")
    validate_type(data, list, label="'variants'")
    validate_not_empty(data, label="'variants'")

    keywords = {
        "identifier", "install-location", "command", "environ", "requirements"
    }

    for index, data in enumerate(data or []):
        validate_type(data, dict, label="'variants/{}'".format(index))
        validate_keywords(data, keywords, label="'variants/{}'".format(index))

        validate_identifier_keyword(data, variant_index=index)
        validate_install_location_keyword(data, variant_index=index)
        validate_command_keyword(data, variant_index=index)
        validate_environ_keyword(data, variant_index=index)
        validate_requirements_keyword(data, variant_index=index)


def validate_keywords(data, keywords, label="Data"):
    """Ensure that no invalid keywords are in *data* mapping.

    :param data: Mapping to validate.

    :param keywords: Set of authorized keywords.

    :param label: String to describe *data* if an error is raised. Default is
        "Data".

    :raise: :exc:`ValueError` if the *data* mapping is incorrect.

    """
    remaining_keywords = set(data.keys()).difference(keywords)
    if remaining_keywords != set():
        raise ValueError(
            "{} contains invalid keywords: {}"
            .format(label, ", ".join(remaining_keywords))
        )


def validate_required(data, label="Data"):
    """Ensure that *data* exists.

    :param data: Content to validate.

    :param label: String to describe *data* if an error is raised. Default is
        "Data".

    :raise: :exc:`ValueError` if *data* is incorrect.

    """
    if data is None:
        raise ValueError("{} is required.".format(label))


def validate_type(data, data_type, label="Data"):
    """Ensure that *data* has correct type.

    :param data: Content to validate.

    :param data_type: Type expected for *data*. It can be a tuple if several
        types are authorized.

    :param label: String to describe *data* if an error is raised. Default is
        "Data".

    :raise: :exc:`ValueError` if *data* is incorrect.

    """
    if data is not None and not isinstance(data, data_type):
        raise ValueError("{} has incorrect type.".format(label))


def validate_not_empty(data, label="Data"):
    """Ensure that *data* container is not empty.

    :param data: Content to validate. It could be a list or a mapping.

    :param label: String to describe *data* if an error is raised. Default is
        "Data".

    :raise: :exc:`ValueError` if *data* is incorrect.

    """
    if data is not None and not len(data):
        raise ValueError("{} should not be empty.".format(label))



# :coding: utf-8

import packaging.requirements
from packaging.requirements import (
    L, Combine, Word, ZeroOrMore, ALPHANUM, Optional, EXTRAS,
    URL_AND_MARKER, VERSION_AND_MARKER, stringStart, stringEnd
)
from packaging.requirements import Requirement

import wiz.symbol

# Extend requirement's expression to allow namespace separators as a valid
# identifier (e.g. "foo::test", "::test").
PUNCTUATION = Word("-_.")
NAME = ALPHANUM + ZeroOrMore(ALPHANUM | (ZeroOrMore(PUNCTUATION) + ALPHANUM))
NAME_SEPARATOR = L(wiz.symbol.NAMESPACE_SEPARATOR)
IDENTIFIER = Combine(
    ZeroOrMore(Optional(NAME) + NAME_SEPARATOR) + NAME
)

packaging.requirements.REQUIREMENT = (
    stringStart + IDENTIFIER("name") + Optional(EXTRAS)
    + (URL_AND_MARKER | VERSION_AND_MARKER) + stringEnd
)


def _display_requirement(_requirement):
    """Improve readability when displaying Requirement instance.

    Before::

        >>> Requirement("nuke>=10,<11")
        <Requirement('nuke<11,>=10')>

    After::

        >>> Requirement("nuke>=10,<11")
        <Requirement('nuke >=10, <11')>

    """
    content = _requirement.name

    variant = ",".join(_requirement.extras)
    if len(variant) > 0:
        content += "[{}]".format(variant)

    if len(_requirement.specifier) > 0:
        content += " " + ", ".join(sorted([
            str(specifier) for specifier in _requirement.specifier
        ], reverse=True))

    return content


#: Monkeypatch magic method to improve readability of serialized item.
Requirement.__str__ = _display_requirement

#: Monkeypatch magic methods to allow comparison between instances.
Requirement.__eq__ = lambda self, other: (
    str(self) == str(other) if isinstance(other, Requirement) else False
)
Requirement.__ne__ = lambda self, other: not (self == other)
Requirement.__hash__ = lambda self: hash(str(self))
